from __future__ import annotations

import importlib

from qadence.logger import get_logger

logger = get_logger(__name__)

try:
    module = importlib.import_module("qadence_protocols.protocols")
    available_protocols = getattr(module, "available_protocols")
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "The 'qadence_protocols' module is not present."
        " Please install the 'qadence-protocol' package."
    )



from __future__ import annotations

import random
from functools import singledispatch
from typing import List

import torch
from numpy.typing import ArrayLike
from torch import Tensor, concat
from torch.distributions import Categorical, Distribution

from qadence.blocks import ChainBlock, KronBlock, PrimitiveBlock, chain, kron
from qadence.circuit import QuantumCircuit
from qadence.execution import run
from qadence.logger import get_script_logger
from qadence.operations import CNOT, RX, RY, RZ, H, I, X
from qadence.types import PI, BackendName, Endianness, StateGeneratorType
from qadence.utils import basis_to_int

# Modules to be automatically added to the qadence namespace
__all__ = [
    "uniform_state",
    "zero_state",
    "one_state",
    "product_state",
    "rand_product_state",
    "ghz_state",
    "random_state",
    "uniform_block",
    "one_block",
    "zero_block",
    "product_block",
    "rand_product_block",
    "ghz_block",
    "pmf",
    "normalize",
    "is_normalized",
    "rand_bitstring",
    "equivalent_state",
]

ATOL_64 = 1e-14  # 64 bit precision
NORMALIZATION_ATOL = ATOL_64
DTYPE = torch.cdouble

parametric_single_qubit_gates: List = [RX, RY, RZ]

logger = get_script_logger(__name__)
# PRIVATE


def _rand_haar_fast(n_qubits: int) -> Tensor:
    # inspired by https://qiskit.org/documentation/_modules/qiskit/quantum_info/states/random.html#random_statevector
    N = 2**n_qubits
    x = -torch.log(torch.rand(N))
    sumx = torch.sum(x)
    phases = torch.rand(N) * 2.0 * PI
    return (torch.sqrt(x / sumx) * torch.exp(1j * phases)).reshape(1, N)


def _rand_haar_slow(n_qubits: int) -> Tensor:
    """
    Detailed in https://arxiv.org/pdf/math-ph/0609050.pdf.

    Textbook implementation, but very expensive. For 12 qubits it takes several seconds.
    For 1 qubit it seems to produce the same distribution as the measure above.
    """
    N = 2**n_qubits
    A = torch.zeros(N, N, dtype=DTYPE).normal_(0, 1)
    B = torch.zeros(N, N, dtype=DTYPE).normal_(0, 1)
    Z = A + 1.0j * B
    Q, R = torch.linalg.qr(Z)
    Lambda = torch.diag(torch.diag(R) / torch.diag(R).abs())
    haar_unitary = torch.matmul(Q, Lambda)
    return torch.matmul(haar_unitary, zero_state(n_qubits).squeeze(0)).unsqueeze(0)


def _from_op(op: type[PrimitiveBlock], n_qubits: int) -> KronBlock:
    return kron(op(i) for i in range(n_qubits))  # type: ignore[arg-type]


def _block_from_bitstring(bitstring: str) -> KronBlock:
    n_qubits = len(bitstring)
    gates = []
    for i, b in zip(range(n_qubits), bitstring):
        gates.append(X(i)) if b == "1" else gates.append(I(i))  # type: ignore[arg-type]
    return kron(*gates)


def _state_from_bitstring(
    bitstring: str, batch_size: int, endianness: Endianness = Endianness.BIG
) -> Tensor:
    n_qubits = len(bitstring)
    wf_batch = torch.zeros(batch_size, 2**n_qubits, dtype=DTYPE)
    k = basis_to_int(basis=bitstring, endianness=endianness)
    wf_batch[:, k] = torch.tensor(1.0 + 0j, dtype=DTYPE)
    return wf_batch


def _abstract_random_state(
    n_qubits: int, batch_size: int = 1
) -> QuantumCircuit | list[QuantumCircuit]:
    qc_list = []
    for i in range(batch_size):
        gates_list = []
        for i in range(n_qubits):
            gate = parametric_single_qubit_gates[
                random.randrange(len(parametric_single_qubit_gates))
            ]
            angle = random.uniform(-2, 2)
            gates_list.append(gate(i, angle))
        qc_list.append(QuantumCircuit(n_qubits, chain(*gates_list)))
    return qc_list[0] if batch_size == 1 else qc_list


# STATES


def uniform_state(n_qubits: int, batch_size: int = 1) -> Tensor:
    """
    Generates the uniform state for a specified number of qubits.

    Arguments:
        n_qubits (int): The number of qubits.
        batch_size (int): The batch size.

    Returns:
        A torch.Tensor.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.states import uniform_state

    state = uniform_state(n_qubits=2)
    print(state)
    ```
    """
    norm = 1 / torch.sqrt(torch.tensor(2**n_qubits))
    return norm * torch.ones(batch_size, 2**n_qubits, dtype=DTYPE)


def zero_state(n_qubits: int, batch_size: int = 1) -> Tensor:
    """
    Generates the zero state for a specified number of qubits.

    Arguments:
        n_qubits (int): The number of qubits for which the zero state is to be generated.
        batch_size (int): The batch size for the zero state.

    Returns:
        A torch.Tensor.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.states import zero_state

    state = zero_state(n_qubits=2)
    print(state)
    ```
    """
    bitstring = "0" * n_qubits
    return _state_from_bitstring(bitstring, batch_size)


def one_state(n_qubits: int, batch_size: int = 1) -> Tensor:
    """
    Generates the one state for a specified number of qubits.

    Arguments:
        n_qubits (int): The number of qubits.
        batch_size (int): The batch size.

    Returns:
        A torch.Tensor.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.states import one_state

    state = one_state(n_qubits=2)
    print(state)
    ```
    """
    bitstring = "1" * n_qubits
    return _state_from_bitstring(bitstring, batch_size)


@singledispatch
def product_state(
    bitstring: str,
    batch_size: int = 1,
    endianness: Endianness = Endianness.BIG,
    backend: BackendName = BackendName.PYQTORCH,
) -> ArrayLike:
    """
    Creates a product state from a bitstring.

    Arguments:
        bitstring (str): A bitstring.
        batch_size (int) : Batch size.
        backend (BackendName): The backend to use. Default is "pyqtorch".

    Returns:
        A torch.Tensor.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.states import product_state

    print(product_state("1100", backend="pyqtorch"))
    print(product_state("1100", backend="horqrux"))
    ```
    """
    if batch_size:
        logger.debug(
            "The input `batch_size` is going to be deprecated. "
            "For now, default batch_size is set to 1."
        )
    return run(product_block(bitstring), backend=backend, endianness=endianness)


@product_state.register
def _(bitstrings: list) -> Tensor:  # type: ignore
    return concat(tuple(product_state(b) for b in bitstrings), dim=0)


def rand_product_state(n_qubits: int, batch_size: int = 1) -> Tensor:
    """
    Creates a random product state.

    Arguments:
        n_qubits (int): The number of qubits.
        batch_size (int): How many bitstrings to use.

    Returns:
        A torch.Tensor.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.states import rand_product_state

    print(rand_product_state(n_qubits=2, batch_size=2))
    ```
    """
    wf_batch = torch.zeros(batch_size, 2**n_qubits, dtype=DTYPE)
    rand_pos = torch.randint(0, 2**n_qubits, (batch_size,))
    wf_batch[torch.arange(batch_size), rand_pos] = torch.tensor(1.0 + 0j, dtype=DTYPE)
    return wf_batch


def ghz_state(n_qubits: int, batch_size: int = 1) -> Tensor:
    """
    Creates a GHZ state.

    Arguments:
        n_qubits (int): The number of qubits.
        batch_size (int): How many bitstrings to use.

    Returns:
        A torch.Tensor.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.states import ghz_state

    print(ghz_state(n_qubits=2, batch_size=2))
    ```
    """
    norm = 1 / torch.sqrt(torch.tensor(2))
    return norm * (zero_state(n_qubits, batch_size) + one_state(n_qubits, batch_size))


def random_state(
    n_qubits: int,
    batch_size: int = 1,
    backend: str = BackendName.PYQTORCH,
    type: StateGeneratorType = StateGeneratorType.HAAR_MEASURE_FAST,
) -> Tensor:
    """
    Generates a random state for a specified number of qubits.

    Arguments:
        n_qubits (int): The number of qubits.
        backend (str): The backend to use.
        batch_size (int): The batch size.
        type : StateGeneratorType.

    Returns:
        A torch.Tensor.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.states import random_state, StateGeneratorType
    from qadence.states import random_state, is_normalized, pmf
    from qadence.types import BackendName
    from torch.distributions import Distribution

    ### We have the following options:
    print([g.value for g in StateGeneratorType])

    n_qubits = 2
    # The default is StateGeneratorType.HAAR_MEASURE_FAST
    state = random_state(n_qubits=n_qubits)
    print(state)

    ### Lets initialize a state using random rotations, i.e., StateGeneratorType.RANDOM_ROTATIONS.
    random = random_state(n_qubits=n_qubits, type=StateGeneratorType.RANDOM_ROTATIONS)
    print(random)
    ```
    """

    if type == StateGeneratorType.HAAR_MEASURE_FAST:
        state = concat(tuple(_rand_haar_fast(n_qubits) for _ in range(batch_size)), dim=0)
    elif type == StateGeneratorType.HAAR_MEASURE_SLOW:
        state = concat(tuple(_rand_haar_slow(n_qubits) for _ in range(batch_size)), dim=0)
    elif type == StateGeneratorType.RANDOM_ROTATIONS:
        state = run(_abstract_random_state(n_qubits, batch_size))  # type: ignore
    assert all(list(map(is_normalized, state)))
    return state


# BLOCKS


def uniform_block(n_qubits: int) -> KronBlock:
    """
    Generates the abstract uniform state for a specified number of qubits.

    Arguments:
        n_qubits (int): The number of qubits.

    Returns:
        A KronBlock representing the uniform state.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.states import uniform_block

    block = uniform_block(n_qubits=2)
    print(block)
    ```
    """
    return _from_op(H, n_qubits=n_qubits)


def one_block(n_qubits: int) -> KronBlock:
    """
    Generates the abstract one state for a specified number of qubits.

    Arguments:
        n_qubits (int): The number of qubits.

    Returns:
        A KronBlock representing the one state.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.states import one_block

    block = one_block(n_qubits=2)
    print(block)
    ```
    """
    return _from_op(X, n_qubits=n_qubits)


def zero_block(n_qubits: int) -> KronBlock:
    """
    Generates the abstract zero state for a specified number of qubits.

    Arguments:
        n_qubits (int): The number of qubits.

    Returns:
        A KronBlock representing the zero state.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.states import zero_block

    block = zero_block(n_qubits=2)
    print(block)
    ```
    """
    return _from_op(I, n_qubits=n_qubits)


def product_block(bitstring: str) -> KronBlock:
    """
    Creates an abstract product state from a bitstring.

    Arguments:
        bitstring (str): A bitstring.

    Returns:
        A KronBlock representing the product state.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.states import product_block

    print(product_block("1100"))
    ```
    """
    return _block_from_bitstring(bitstring)


def rand_product_block(n_qubits: int) -> KronBlock:
    """
    Creates a block representing a random abstract product state.

    Arguments:
        n_qubits (int): The number of qubits.

    Returns:
        A KronBlock representing the product state.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.states import rand_product_block

    print(rand_product_block(n_qubits=2))
    ```
    """
    return product_block(rand_bitstring(n_qubits))


def ghz_block(n_qubits: int) -> ChainBlock:
    """
    Generates the abstract ghz state for a specified number of qubits.

    Arguments:
        n_qubits (int): The number of qubits.

    Returns:
        A ChainBlock representing the GHZ state.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.states import ghz_block

    block = ghz_block(n_qubits=2)
    print(block)
    ```
    """
    cnots = chain(CNOT(i - 1, i) for i in range(1, n_qubits))
    return chain(H(0), cnots)


# UTILITIES


def pmf(wf: Tensor) -> Distribution:
    """
    Converts a wave function into a torch Distribution.

    Arguments:
        wf (torch.Tensor): The wave function as a torch tensor.

    Returns:
        A torch.distributions.Distribution.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.states import uniform_state, pmf

    print(pmf(uniform_state(2)).probs)
    ```
    """
    return Categorical(torch.abs(torch.pow(wf, 2)))


def normalize(wf: Tensor) -> Tensor:
    """
    Normalizes a wavefunction or batch of wave functions.

    Arguments:
        wf (torch.Tensor): Normalized wavefunctions.

    Returns:
        A torch.Tensor.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.states import uniform_state, normalize

    print(normalize(uniform_state(2, 2)))
    ```
    """
    if wf.dim() == 1:
        return wf / torch.sqrt((wf.abs() ** 2).sum())
    else:
        return wf / torch.sqrt((wf.abs() ** 2).sum(1)).unsqueeze(1)


def is_normalized(wf: Tensor, atol: float = NORMALIZATION_ATOL) -> bool:
    """
    Checks if a wave function is normalized.

    Arguments:
        wf (torch.Tensor): The wave function as a torch tensor.
        atol (float) : The tolerance.

    Returns:
        A bool.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.states import uniform_state, is_normalized

    print(is_normalized(uniform_state(2)))
    ```
    """
    if wf.dim() == 1:
        wf = wf.unsqueeze(0)
    sum_probs: Tensor = (wf.abs() ** 2).sum(dim=1)
    ones = torch.ones_like(sum_probs)
    return torch.allclose(sum_probs, ones, rtol=0.0, atol=atol)  # type: ignore[no-any-return]


def rand_bitstring(N: int) -> str:
    """
    Creates a random bistring.

    Arguments:
        N (int): The length of the bitstring.

    Returns:
        A string.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.states import rand_bitstring

    print(rand_bitstring(N=8))
    ```
    """
    return "".join(str(random.randint(0, 1)) for _ in range(N))


def equivalent_state(
    s0: torch.Tensor, s1: torch.Tensor, rtol: float = 0.0, atol: float = NORMALIZATION_ATOL
) -> bool:
    from qadence.overlap import fidelity

    fid = fidelity(s0, s1)
    expected = torch.ones_like(fid)
    return torch.allclose(fid, expected, rtol=rtol, atol=atol)  # type: ignore[no-any-return]



from __future__ import annotations

from math import log2
from typing import Any

import jax.numpy as jnp
from jax import Array, device_get
from sympy import Expr
from sympy2jax import SymbolicModule as JaxSympyModule
from torch import Tensor, cdouble, from_numpy

from qadence.blocks import (
    AbstractBlock,
    AddBlock,
    ChainBlock,
    KronBlock,
    ProjectorBlock,
    ScaleBlock,
)
from qadence.blocks.block_to_tensor import _gate_parameters
from qadence.types import Endianness, ParamDictType


def jarr_to_tensor(arr: Array, dtype: Any = cdouble) -> Tensor:
    return from_numpy(device_get(arr)).to(dtype=dtype)


def tensor_to_jnp(tensor: Tensor, dtype: Any = jnp.complex128) -> Array:
    return (
        jnp.array(tensor.numpy(), dtype=dtype)
        if not tensor.requires_grad
        else jnp.array(tensor.detach().numpy(), dtype=dtype)
    )


def values_to_jax(param_values: dict[str, Tensor]) -> dict[str, Array]:
    return {key: jnp.array(value.detach().numpy()) for key, value in param_values.items()}


def jaxify(expr: Expr) -> JaxSympyModule:
    return JaxSympyModule(expr)


def unhorqify(state: Array) -> Array:
    """Convert a state of shape [2] * n_qubits + [batch_size] to (batch_size, 2**n_qubits)."""
    return jnp.ravel(state)


def horqify(state: Array) -> Array:
    n_qubits = int(log2(state.shape[1]))
    return state.reshape([2] * n_qubits)


def uniform_batchsize(param_values: ParamDictType) -> ParamDictType:
    max_batch_size = max(p.size for p in param_values.values())
    batched_values = {
        k: (v if v.size == max_batch_size else v.repeat(max_batch_size))
        for k, v in param_values.items()
    }
    return batched_values


IMAT = jnp.eye(2, dtype=jnp.cdouble)
ZEROMAT = jnp.zeros_like(IMAT)
XMAT = jnp.array([[0, 1], [1, 0]], dtype=jnp.cdouble)
YMAT = jnp.array([[0, -1j], [1j, 0]], dtype=jnp.cdouble)
ZMAT = jnp.array([[1, 0], [0, -1]], dtype=jnp.cdouble)

MAT_DICT = {"I": IMAT, "Z": ZMAT, "Y": YMAT, "X": XMAT}


def _fill_identities(
    block_mat: Array,
    qubit_support: tuple,
    full_qubit_support: tuple | list,
    diag_only: bool = False,
    endianness: Endianness = Endianness.BIG,
) -> Array:
    qubit_support = tuple(sorted(qubit_support))
    mat = IMAT if qubit_support[0] != full_qubit_support[0] else block_mat
    if diag_only:
        mat = jnp.diag(mat.squeeze(0))
    for i in full_qubit_support[1:]:
        if i == qubit_support[0]:
            other = jnp.diag(block_mat) if diag_only else block_mat
            mat = jnp.kron(mat, other)
        elif i not in qubit_support:
            other = jnp.diag(IMAT) if diag_only else IMAT
            mat = jnp.kron(mat, other)
    return mat


def block_to_jax(
    block: AbstractBlock,
    values: dict = None,
    qubit_support: tuple | None = None,
    use_full_support: bool = True,
    endianness: Endianness = Endianness.BIG,
) -> Array:
    if values is None:
        from qadence.blocks import embedding

        (ps, embed) = embedding(block)
        values = embed(ps, {})

    # get number of qubits
    if qubit_support is None:
        if use_full_support:
            qubit_support = tuple(range(0, block.n_qubits))
        else:
            qubit_support = block.qubit_support
    nqubits = len(qubit_support)

    if isinstance(block, (ChainBlock, KronBlock)):
        # create identity matrix of appropriate dimensions
        mat = IMAT
        for i in range(nqubits - 1):
            mat = jnp.kron(mat, IMAT)

        # perform matrix multiplications
        for b in block.blocks:
            other = block_to_jax(b, values, qubit_support, endianness=endianness)
            mat = jnp.matmul(other, mat)

    elif isinstance(block, AddBlock):
        # create zero matrix of appropriate dimensions
        mat = ZEROMAT
        for _ in range(nqubits - 1):
            mat = jnp.kron(mat, ZEROMAT)

        # perform matrix summation
        for b in block.blocks:
            mat = mat + block_to_jax(b, values, qubit_support, endianness=endianness)

    elif isinstance(block, ScaleBlock):
        (scale,) = _gate_parameters(block, values)
        if isinstance(scale, Tensor):
            scale = tensor_to_jnp(scale)
        mat = scale * block_to_jax(block.block, values, qubit_support, endianness=endianness)

    elif block.name in MAT_DICT.keys():
        block_mat = MAT_DICT[block.name]

        # add missing identities on unused qubits
        mat = _fill_identities(block_mat, block.qubit_support, qubit_support, endianness=endianness)

    elif isinstance(block, ProjectorBlock):
        from qadence.states import product_state

        bra = tensor_to_jnp(product_state(block.bra))
        ket = tensor_to_jnp(product_state(block.ket))

        block_mat = jnp.kron(ket, bra.T)

        mat = _fill_identities(
            block_mat,
            block.qubit_support,
            qubit_support,
            endianness=endianness,
        )
    else:
        raise TypeError(f"Conversion for block type {type(block)} not supported.")

    return mat



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from pasqal_cloud import TokenProvider
from pasqal_cloud.device import EmulatorType
from pulser_simulation.simconfig import SimConfig

from qadence.backend import BackendConfiguration

DEFAULT_CLOUD_ENV = "prod"


@dataclass
class CloudConfiguration:
    platform: EmulatorType = EmulatorType.EMU_FREE
    username: str | None = None
    password: str | None = None
    project_id: str | None = None
    environment: str = "prod"
    token_provider: TokenProvider | None = None


@dataclass
class Configuration(BackendConfiguration):
    sampling_rate: float = 1.0
    """Sampling rate to be used for local simulations.

    Set to 1.0
    to avoid any interpolation in the solving procedure
    """

    method_solv: str = "adams"
    """Solver method to pass to the Qutip solver."""

    n_steps_solv: float = 1e8
    """Number of solver steps to pass to the Qutip solver."""

    sim_config: Optional[SimConfig] = None
    """Simulation configuration with optional noise options."""

    with_modulation: bool = False
    """Add laser modulation to the local execution.

    This will take
    into account realistic laser pulse modulation when simulating
    a pulse sequence
    """

    amplitude_local: Optional[float] = None
    """Default pulse amplitude on local channel.

    FIXME: To be deprecated.
    """

    amplitude_global: Optional[float] = None
    """Default pulse amplitude on global channel.

    FIXME: To be deprecated.
    """

    detuning: Optional[float] = None
    """Default value for the detuning pulses.

    FIXME: To be deprecated.
    """

    # configuration for cloud simulations
    cloud_configuration: Optional[CloudConfiguration] = None

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.sim_config is not None and not isinstance(self.sim_config, SimConfig):
            raise TypeError("Wrong 'sim_config' attribute type, pass a valid SimConfig object!")

        if isinstance(self.cloud_configuration, dict):
            self.cloud_configuration = CloudConfiguration(**self.cloud_configuration)



from __future__ import annotations

from logging import getLogger

import sympy
import torch
from rich.console import Console, RenderableType
from rich.panel import Panel
from rich.tree import Tree
from torch import Tensor, cdouble

from qadence.blocks import (
    ControlBlock,
    ParametricControlBlock,
)
from qadence.blocks.utils import (
    add,  # noqa
    chain,
    kron,
)
from qadence.parameters import (
    Parameter,
    evaluate,
)
from qadence.types import OpName, TNumber, TParameter

from .parametric import PHASE, RX, RY, RZ
from .primitive import SWAP, I, N, X, Y, Z

logger = getLogger(__name__)


class CNOT(ControlBlock):
    """The CNot, or CX, gate."""

    name = OpName.CNOT

    def __init__(self, control: int, target: int) -> None:
        self.generator = kron(N(control), X(target) - I(target))
        super().__init__((control,), X(target))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-2, 0, 0, 0], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor([-1, 1, 1, 1], dtype=cdouble)

    def __ascii__(self, console: Console) -> RenderableType:
        (target, control) = self.qubit_support
        h = abs(target - control) + 1
        return Panel(self._block_title, expand=False, height=3 * h)

    def __rich_tree__(self, tree: Tree = None) -> Tree:
        if tree is None:
            return Tree(self._block_title)
        else:
            tree.add(self._block_title)
        return tree


class MCZ(ControlBlock):
    name = OpName.MCZ

    def __init__(self, control: tuple[int, ...], target: int) -> None:
        self.generator = kron(*[N(qubit) for qubit in control], Z(target) - I(target))
        super().__init__(control, Z(target))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.cat((torch.tensor(-2, dtype=cdouble), torch.zeros(2**self.n_qubits - 1)))

    @property
    def eigenvalues(self) -> Tensor:
        torch.cat((torch.tensor(-1, dtype=cdouble), torch.ones(2**self.n_qubits - 1)))

    def __ascii__(self, console: Console) -> RenderableType:
        (target, control) = self.qubit_support
        h = abs(target - control) + 1
        return Panel(self._block_title, expand=False, height=3 * h)

    def __rich_tree__(self, tree: Tree = None) -> Tree:
        if tree is None:
            return Tree(self._block_title)
        else:
            tree.add(self._block_title)
        return tree


class CZ(MCZ):
    """The CZ gate."""

    name = OpName.CZ

    def __init__(self, control: int, target: int) -> None:
        super().__init__((control,), target)


class MCRX(ParametricControlBlock):
    name = OpName.MCRX

    def __init__(
        self,
        control: tuple[int, ...],
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ) -> None:
        self.generator = kron(*[N(qubit) for qubit in control], X(target))
        super().__init__(control, RX(target, parameter))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.cat(
            (torch.zeros(2**self.n_qubits - 2), torch.tensor([1, -1], dtype=cdouble))
        )

    @property
    def eigenvalues(self) -> Tensor:
        val = evaluate(self.parameters.parameter, as_torch=True)
        lmbd = torch.cos(val / 2.0) - 1j * torch.sin(val / 2.0)
        return torch.cat((torch.ones(2**self.n_qubits - 2), lmbd, lmbd.conj()))


class CRX(MCRX):
    """The CRX gate."""

    name = OpName.CRX

    def __init__(
        self,
        control: int,
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ):
        super().__init__((control,), target, parameter)


class MCRY(ParametricControlBlock):
    name = OpName.MCRY

    def __init__(
        self,
        control: tuple[int, ...],
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ) -> None:
        self.generator = kron(*[N(qubit) for qubit in control], Y(target))
        super().__init__(control, RY(target, parameter))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.cat(
            (torch.zeros(2**self.n_qubits - 2), torch.tensor([1, -1], dtype=cdouble))
        )

    @property
    def eigenvalues(self) -> Tensor:
        val = evaluate(self.parameters.parameter, as_torch=True)
        lmbd = torch.cos(val / 2.0) - 1j * torch.sin(val / 2.0)
        return torch.cat((torch.ones(2**self.n_qubits - 2), lmbd, lmbd.conj()))


class CRY(MCRY):
    """The CRY gate."""

    name = OpName.CRY

    def __init__(
        self,
        control: int,
        target: int,
        parameter: TParameter,
    ):
        super().__init__((control,), target, parameter)


class MCRZ(ParametricControlBlock):
    name = OpName.MCRZ

    def __init__(
        self,
        control: tuple[int, ...],
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ) -> None:
        self.generator = kron(*[N(qubit) for qubit in control], Z(target))
        super().__init__(control, RZ(target, parameter))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.cat(
            (torch.zeros(2**self.n_qubits - 2), torch.tensor([1, -1], dtype=cdouble))
        )

    @property
    def eigenvalues(self) -> Tensor:
        val = evaluate(self.parameters.parameter, as_torch=True)
        lmbd = torch.cos(val / 2.0) - 1j * torch.sin(val / 2.0)
        return torch.cat((torch.ones(2**self.n_qubits - 2), lmbd, lmbd.conj()))


class CRZ(MCRZ):
    """The CRZ gate."""

    name = OpName.CRZ

    def __init__(
        self,
        control: int,
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ):
        super().__init__((control,), target, parameter)


class MCPHASE(ParametricControlBlock):
    name = OpName.MCPHASE

    def __init__(
        self,
        control: tuple[int, ...],
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ) -> None:
        self.generator = kron(*[N(qubit) for qubit in control], Z(target) - I(target))
        super().__init__(control, PHASE(target, parameter))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.cat(
            (torch.tensor([-2, 0], dtype=cdouble), (torch.zeros(2**self.n_qubits - 2)))
        )

    @property
    def eigenvalues(self) -> Tensor:
        v = evaluate(self.parameters.parameter, as_torch=True)
        return torch.cat((torch.ones(2**self.n_qubits - 1), torch.exp(1j * v)))

    def __rich_tree__(self, tree: Tree = None) -> Tree:
        if tree is None:
            return Tree(self._block_title)
        else:
            tree.add(self._block_title)
        return tree

    def __ascii__(self, console: Console) -> RenderableType:
        (target, control) = self.qubit_support
        h = abs(target - control) + 1
        return Panel(self._block_title, expand=False, height=3 * h)


class CPHASE(MCPHASE):
    """The CPHASE gate."""

    name = OpName.CPHASE

    def __init__(
        self,
        control: int,
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ):
        super().__init__((control,), target, parameter)


class CSWAP(ControlBlock):
    """The CSWAP (Control-SWAP) gate."""

    name = OpName.CSWAP

    def __init__(self, control: int | tuple[int, ...], target1: int, target2: int) -> None:
        if isinstance(control, tuple):
            control = control[0]

        a00m = -N(target=control)
        a00p = -N(target=control, state="0")
        a11 = -N(target=target1)
        a22 = -N(target=target2, state="0")
        a12 = 0.5 * (chain(X(target1), Z(target1)) + X(target1))
        a21 = 0.5 * (chain(Z(target2), X(target2)) + X(target2))
        no_effect = kron(a00m, I(target1), I(target2))
        swap_effect = (
            kron(a00p, -1.0 * a22, a11)
            + kron(a00p, -1.0 * a11, a22)
            + kron(a00p, a12, a21)
            + kron(a00p, a21, a12)
        )
        self.generator = no_effect + swap_effect
        super().__init__((control,), SWAP(target1, target2))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor((1, -1, 1, 1, 1, 1, 1, 1), dtype=torch.cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor((1, -1, 1, 1, 1, 1, 1, 1), dtype=torch.cdouble)

    @property
    def nqubits(self) -> int:
        return 3


class Toffoli(ControlBlock):
    name = OpName.TOFFOLI

    def __init__(self, control: tuple[int, ...], target: int) -> None:
        self.generator = kron(*[N(qubit) for qubit in control], X(target) - I(target))
        super().__init__(control, X(target))

    @property
    def n_qubits(self) -> int:
        return len(self.qubit_support)

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor(
            [-2, *[0 for _ in range(2 ** len(self.qubit_support) - 1)]], dtype=cdouble
        )

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor(
            [-1, *[1 for _ in range(2 ** len(self.qubit_support) - 1)]], dtype=cdouble
        )



from __future__ import annotations

from dataclasses import dataclass, field
from functools import reduce
from itertools import chain as flatten
from operator import add
from typing import Any, Callable, Dict

import jax.numpy as jnp
from horqrux.abstract import Primitive as Gate
from horqrux.analog import _HamiltonianEvolution as NativeHorqHEvo
from horqrux.apply import apply_gate
from horqrux.parametric import RX, RY, RZ
from horqrux.primitive import NOT, SWAP, H, I, X, Y, Z
from horqrux.utils import inner
from jax import Array
from jax.scipy.linalg import expm
from jax.tree_util import register_pytree_node_class

from qadence.backends.jax_utils import block_to_jax
from qadence.blocks import (
    AbstractBlock,
    AddBlock,
    CompositeBlock,
    ParametricBlock,
    PrimitiveBlock,
    ScaleBlock,
    TimeEvolutionBlock,
)
from qadence.operations import (
    CNOT,
    CRX,
    CRY,
    CRZ,
    MCRX,
    MCRY,
    MCRZ,
    MCZ,
)
from qadence.operations import SWAP as QDSWAP
from qadence.types import OpName, ParamDictType

from .config import Configuration

ops_map: Dict[str, Callable] = {
    OpName.X: X,
    OpName.Y: Y,
    OpName.Z: Z,
    OpName.H: H,
    OpName.RX: RX,
    OpName.RY: RY,
    OpName.RZ: RZ,
    OpName.CRX: RX,
    OpName.CRY: RY,
    OpName.CRZ: RZ,
    OpName.CNOT: NOT,
    OpName.I: I,
    OpName.SWAP: SWAP,
}

supported_gates = list(set(list(ops_map.keys())))


@register_pytree_node_class
@dataclass
class HorqruxCircuit:
    operators: list[Gate] = field(default_factory=list)

    def tree_flatten(self) -> tuple[tuple[list[Any]], tuple[()]]:
        children = (self.operators,)
        aux_data = ()
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data: Any, children: Any) -> Any:
        return cls(*children, *aux_data)

    def _forward(self, state: Array, values: ParamDictType) -> Array:
        return reduce(lambda state, gate: gate.forward(state, values), self.operators, state)

    def forward(self, state: Array, values: ParamDictType) -> Array:
        return self._forward(state, values)


@register_pytree_node_class
@dataclass
class HorqruxObservable(HorqruxCircuit):
    def __init__(self, operators: list[Gate]):
        super().__init__(operators=operators)

    def forward(self, state: Array, values: ParamDictType) -> Array:
        return jnp.real(inner(state, self._forward(state, values)))


def convert_observable(
    block: AbstractBlock, n_qubits: int, config: Configuration
) -> HorqruxObservable:
    ops = convert_block(block, n_qubits, config)
    return HorqruxObservable(ops)


def convert_block(
    block: AbstractBlock, n_qubits: int = None, config: Configuration = Configuration()
) -> list:
    if n_qubits is None:
        n_qubits = max(block.qubit_support) + 1
    ops = []
    if isinstance(block, CompositeBlock):
        ops = list(flatten(*(convert_block(b, n_qubits, config) for b in block.blocks)))
        ops = [HorqAddGate(ops)] if isinstance(block, AddBlock) else [HorqruxCircuit(ops)]
    elif isinstance(block, ScaleBlock):
        op = convert_block(block.block, n_qubits, config=config)[0]
        param_name = config.get_param_name(block)[0]
        ops = [HorqScaleGate(op, param_name)]
    elif block.name in ops_map.keys():
        native_op_fn = ops_map[block.name]
        target, control = (
            (block.qubit_support[1], block.qubit_support[0])
            if isinstance(block, (CNOT, CRX, CRY, CRZ, QDSWAP))
            else (block.qubit_support[0], (None,))
        )
        native_gate: Gate
        if isinstance(block, ParametricBlock):
            if len(block.parameters._uuid_dict) > 1:
                raise NotImplementedError("Only single-parameter operations are supported.")
            param_name = config.get_param_name(block)[0]
            native_gate = native_op_fn(param=param_name, target=target, control=control)

        elif isinstance(block, PrimitiveBlock):
            if isinstance(block, QDSWAP):
                native_gate = native_op_fn(block.qubit_support[::-1])
            else:
                native_gate = native_op_fn(target=target, control=control)
        ops = [HorqOperation(native_gate)]

    elif isinstance(block, (MCRX, MCRY, MCRZ, MCZ)):
        block_name = block.name[2:] if block.name.startswith("M") else block.name
        native_op_fn = ops_map[block_name]
        control = block.qubit_support[:-1]
        target = block.qubit_support[-1]

        if isinstance(block, ParametricBlock):
            param = config.get_param_name(block)[0]
            native_gate = native_op_fn(param=param, target=target, control=control)
        else:
            native_gate = native_op_fn(target, control)
        ops = [HorqOperation(native_gate)]
    elif isinstance(block, TimeEvolutionBlock):
        ops = [HorqHamiltonianEvolution(block, config)]
    else:
        raise NotImplementedError(f"Non-supported operation of type {type(block)}.")

    return ops


@register_pytree_node_class
class HorqAddGate(HorqruxCircuit):
    def __init__(self, operations: list[Gate]):
        self.operators = operations
        self.name = "Add"

    def forward(self, state: Array, values: ParamDictType = {}) -> Array:
        return reduce(add, (gate.forward(state, values) for gate in self.operators))

    def __repr__(self) -> str:
        return self.name + f"({self.operators})"


@register_pytree_node_class
@dataclass
class HorqOperation:
    def __init__(self, native_gate: Gate):
        self.native_gate = native_gate

    def forward(self, state: Array, values: ParamDictType) -> Array:
        return apply_gate(state, self.native_gate, values)

    def tree_flatten(self) -> tuple[tuple[Gate], tuple[()]]:
        children = (self.native_gate,)
        aux_data = ()
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data: Any, children: Any) -> Any:
        return cls(*children, *aux_data)


@register_pytree_node_class
@dataclass
class HorqScaleGate:
    def __init__(self, gate: HorqOperation, parameter_name: str):
        self.gate = gate
        self.parameter: str = parameter_name

    def forward(self, state: Array, values: ParamDictType) -> Array:
        return jnp.array(values[self.parameter]) * self.gate.forward(state, values)

    def tree_flatten(self) -> tuple[tuple[HorqOperation], tuple[str]]:
        children = (self.gate,)
        aux_data = (self.parameter,)
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data: Any, children: Any) -> Any:
        return cls(*children, *aux_data)


@register_pytree_node_class
@dataclass
class HorqHamiltonianEvolution(NativeHorqHEvo):
    def __init__(
        self,
        block: TimeEvolutionBlock,
        config: Configuration,
    ):
        super().__init__("I", block.qubit_support, (None,))
        self.qubit_support = block.qubit_support
        self.param_names = config.get_param_name(block)
        self.block = block
        self.hmat: Array

        if isinstance(block.generator, AbstractBlock) and not block.generator.is_parametric:
            hmat = block_to_jax(
                block.generator,
                qubit_support=self.qubit_support,
                use_full_support=False,
            )
            self.hmat = hmat
            self._hamiltonian = lambda self, values: self.hmat

        else:

            def _hamiltonian(self: HorqHamiltonianEvolution, values: dict[str, Array]) -> Array:
                hmat = block_to_jax(
                    block.generator,  # type: ignore[arg-type]
                    values=values,
                    qubit_support=self.qubit_support,
                    use_full_support=False,
                )
                return hmat

            self._hamiltonian = _hamiltonian

        self._time_evolution = lambda values: values[self.param_names[0]]

    def unitary(self, values: dict[str, Array]) -> Array:
        """The evolved operator given current parameter values for generator and time evolution."""
        return expm(self._hamiltonian(self, values) * (-1j * self._time_evolution(values)))

    def forward(
        self,
        state: Array,
        values: dict[str, Array],
    ) -> Array:
        return apply_gate(state, self, values)

    def tree_flatten(self) -> tuple[tuple[NativeHorqHEvo], tuple]:
        children = (self,)
        aux_data = ()
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data: Any, children: Any) -> Any:
        return cls(*children, *aux_data)



from __future__ import annotations

import os
from functools import lru_cache
from typing import Optional

from pasqal_cloud import AUTH0_CONFIG, PASQAL_ENDPOINTS, SDK, Auth0Conf, Endpoints, TokenProvider

from .config import DEFAULT_CLOUD_ENV, CloudConfiguration


@lru_cache(maxsize=5)
def _get_client(
    username: Optional[str] = None,
    password: Optional[str] = None,
    project_id: Optional[str] = None,
    environment: Optional[str] = None,
    token_provider: Optional[TokenProvider] = None,
) -> SDK:
    auth0conf: Optional[Auth0Conf] = None
    endpoints: Optional[Endpoints] = None

    username = os.environ.get("PASQAL_CLOUD_USERNAME", "") if username is None else username
    password = os.environ.get("PASQAL_CLOUD_PASSWORD", "") if password is None else password
    project_id = os.environ.get("PASQAL_CLOUD_PROJECT_ID", "") if project_id is None else project_id

    environment = (
        os.environ.get("PASQAL_CLOUD_ENV", DEFAULT_CLOUD_ENV)
        if environment is None
        else environment
    )

    # setup configuration for environments different than production
    if environment == "preprod":
        auth0conf = AUTH0_CONFIG["preprod"]
        endpoints = PASQAL_ENDPOINTS["preprod"]
    elif environment == "dev":
        auth0conf = AUTH0_CONFIG["dev"]
        endpoints = PASQAL_ENDPOINTS["dev"]

    if all([username, password, project_id]) or all([token_provider, project_id]):
        pass
    else:
        raise Exception("You must either provide project_id and log-in details or a token provider")

    return SDK(
        username=username,
        password=password,
        project_id=project_id,
        auth0=auth0conf,
        endpoints=endpoints,
        token_provider=token_provider,
    )


def get_client(credentials: CloudConfiguration) -> SDK:
    return _get_client(
        username=credentials.username,
        password=credentials.password,
        project_id=credentials.project_id,
        environment=credentials.environment,
        token_provider=credentials.token_provider,
    )



from __future__ import annotations

from logging import getLogger
from typing import Union

import numpy as np
import torch
from rich.console import Console
from rich.padding import Padding
from torch import Tensor, cdouble, tensor

from qadence.blocks import (
    AbstractBlock,
    PrimitiveBlock,
)
from qadence.blocks.primitive import ProjectorBlock
from qadence.blocks.utils import (
    chain,
    kron,
)
from qadence.parameters import (
    Parameter,
)
from qadence.types import OpName, TNumber

logger = getLogger(__name__)


class X(PrimitiveBlock):
    """The X gate."""

    name = OpName.X

    def __init__(self, target: int):
        super().__init__((target,))

    @property
    def generator(self) -> AbstractBlock:
        return self

    @property
    def eigenvalues_generator(self) -> Tensor:
        return tensor([-1, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([-1, 1], dtype=cdouble)


class Y(PrimitiveBlock):
    """The Y gate."""

    name = OpName.Y

    def __init__(self, target: int):
        super().__init__((target,))

    @property
    def generator(self) -> AbstractBlock:
        return self

    @property
    def eigenvalues_generator(self) -> Tensor:
        return tensor([-1, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([-1, 1], dtype=cdouble)


class Z(PrimitiveBlock):
    """The Z gate."""

    name = OpName.Z

    def __init__(self, target: int):
        super().__init__((target,))

    @property
    def generator(self) -> AbstractBlock:
        return self

    @property
    def eigenvalues_generator(self) -> Tensor:
        return tensor([-1, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([-1, 1], dtype=cdouble)


class I(PrimitiveBlock):
    """The identity gate."""

    name = OpName.I

    def __init__(self, target: int):
        super().__init__((target,))

    def __ixor__(self, other: AbstractBlock | int) -> AbstractBlock:
        if not isinstance(other, AbstractBlock):
            raise ValueError(
                f"Can only initialize a kron block with another block. Got {type(other)}."
            )
        return other

    def __imul__(self, other: AbstractBlock | TNumber | Parameter) -> AbstractBlock:
        if not isinstance(other, AbstractBlock):
            raise ValueError(
                "In-place multiplication is available " "only for AbstractBlock instances"
            )
        return other

    @property
    def generator(self) -> AbstractBlock:
        return I(*self.qubit_support)

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.ones(2, dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.ones(2, dtype=cdouble)

    def __ascii__(self, console: Console) -> Padding:
        return Padding("──────", (1, 1, 1, 1))


class Projector(ProjectorBlock):
    """The projector operator."""

    name = OpName.PROJ

    def __init__(
        self,
        ket: str,
        bra: str,
        qubit_support: int | tuple[int, ...],
    ):
        super().__init__(ket=ket, bra=bra, qubit_support=qubit_support)

    @property
    def generator(self) -> None:
        raise ValueError("Property `generator` not available for non-unitary operator.")

    @property
    def eigenvalues_generator(self) -> None:
        raise ValueError("Property `eigenvalues_generator` not available for non-unitary operator.")


class N(Projector):
    """The N = (1/2)(I-Z) operator."""

    name = OpName.N

    def __init__(self, target: int, state: str = "1"):
        super().__init__(ket=state, bra=state, qubit_support=(target,))

    @property
    def generator(self) -> None:
        raise ValueError("Property `generator` not available for non-unitary operator.")

    @property
    def eigenvalues_generator(self) -> None:
        raise ValueError("Property `eigenvalues_generator` not available for non-unitary operator.")

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([0, 1], dtype=cdouble)


class S(PrimitiveBlock):
    """The S / Phase gate."""

    name = OpName.S

    def __init__(self, target: int):
        self.generator = I(target) - Z(target)
        super().__init__((target,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return tensor([0, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([1, 1j], dtype=cdouble)

    def dagger(self) -> SDagger:
        return SDagger(*self.qubit_support)


class SDagger(PrimitiveBlock):
    """The Hermitian adjoint/conjugate transpose of the S / Phase gate."""

    name = OpName.SDAGGER

    def __init__(self, target: int):
        self.generator = I(target) - Z(target)
        super().__init__((target,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return tensor([0, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([1, -1j], dtype=cdouble)

    def dagger(self) -> S:
        return S(*self.qubit_support)


class H(PrimitiveBlock):
    """The Hadamard or H gate."""

    name = OpName.H

    def __init__(self, target: int):
        self.generator = (1 / np.sqrt(2)) * (X(target) + Z(target) - np.sqrt(2) * I(target))
        super().__init__((target,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-2, 0], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor([-1, 1], dtype=cdouble)


class Zero(PrimitiveBlock):
    name = OpName.ZERO

    def __init__(self) -> None:
        self.generator = 0 * I(0)
        super().__init__((0,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.zeros(2, dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.zeros(2, dtype=cdouble)

    def __add__(self, other: AbstractBlock) -> AbstractBlock:
        return other

    def __iadd__(self, other: AbstractBlock) -> AbstractBlock:
        return other

    def __sub__(self, other: AbstractBlock) -> AbstractBlock:
        return -other

    def __isub__(self, other: AbstractBlock) -> AbstractBlock:
        return -other

    def __mul__(self, other: AbstractBlock | TNumber | Parameter) -> AbstractBlock:
        return self

    def __imul__(self, other: AbstractBlock | TNumber | Parameter) -> AbstractBlock:
        return self

    def __ixor__(self, other: AbstractBlock | TNumber | Parameter) -> AbstractBlock:
        return self

    def __pow__(self, power: int) -> AbstractBlock:
        return self


class T(PrimitiveBlock):
    """The T gate."""

    name = OpName.T

    def __init__(self, target: int):
        self.generator = I(target) - Z(target)
        super().__init__((target,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([0, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor([1.0, torch.sqrt(torch.tensor([1j]))], dtype=cdouble)

    @property
    def n_qubits(self) -> int:
        return 1

    def dagger(self) -> TDagger:
        return TDagger(*self.qubit_support)


class TDagger(PrimitiveBlock):
    """The Hermitian adjoint/conjugate transpose of the T gate."""

    # FIXME: this gate is not support by any backend
    name = "T_dagger"

    def __init__(self, target: int):
        self.generator = I(target) - Z(target)
        super().__init__((target,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([0, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor([1.0, torch.sqrt(torch.tensor([-1j]))], dtype=cdouble)

    @property
    def n_qubits(self) -> int:
        return 1

    def dagger(self) -> T:
        return T(*self.qubit_support)


class SWAP(PrimitiveBlock):
    """The SWAP gate."""

    name = OpName.SWAP

    def __init__(self, control: int, target: int) -> None:
        a11 = 0.5 * (Z(control) - I(control))
        a22 = -0.5 * (Z(target) + I(target))
        a12 = 0.5 * (chain(X(control), Z(control)) + X(control))
        a21 = 0.5 * (chain(Z(target), X(target)) + X(target))
        self.generator = (
            kron(-1.0 * a22, a11) + kron(-1.0 * a11, a22) + kron(a12, a21) + kron(a21, a12)
        )
        super().__init__((control, target))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-2, 0, 0, 0], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor([-1, 1, 1, 1], dtype=cdouble)

    @property
    def n_qubits(self) -> int:
        return 2

    @property
    def _block_title(self) -> str:
        c, t = self.qubit_support
        s = f"{self.name}({c}, {t})"
        return s if self.tag is None else (s + rf" \[tag: {self.tag}]")


TPauliBlock = Union[X, Y, Z, I, N]



from __future__ import annotations

from collections import Counter
from dataclasses import dataclass, field
from logging import getLogger
from typing import Any

import numpy as np
import torch
from braket.circuits import Circuit as BraketCircuit
from braket.devices import LocalSimulator
from torch import Tensor

from qadence.backend import Backend as BackendInterface
from qadence.backend import ConvertedCircuit, ConvertedObservable
from qadence.backends.utils import to_list_of_dicts
from qadence.blocks import AbstractBlock, block_to_tensor
from qadence.circuit import QuantumCircuit
from qadence.measurements import Measurements
from qadence.mitigations import Mitigations
from qadence.mitigations.protocols import apply_mitigation
from qadence.noise import Noise
from qadence.noise.protocols import apply_noise
from qadence.overlap import overlap_exact
from qadence.transpile import transpile
from qadence.types import BackendName, Engine
from qadence.utils import Endianness

from .config import Configuration, default_passes
from .convert_ops import convert_block

logger = getLogger(__name__)


def promote_parameters(parameters: dict[str, Tensor | float]) -> dict[str, float]:
    float_params = {}
    for name, value in parameters.items():
        try:
            v = value if isinstance(value, float) else value.item()
            float_params[name] = v
        except ValueError:
            raise ValueError("Currently batching is not supported with Braket digital")
    return float_params


@dataclass(frozen=True, eq=True)
class Backend(BackendInterface):
    name: BackendName = BackendName.BRAKET
    supports_ad: bool = False
    supports_adjoint: bool = False
    # TODO Use native braket adjoint differentiation.
    support_bp: bool = False
    is_remote: bool = False
    with_measurements: bool = True
    with_noise: bool = False
    native_endianness: Endianness = Endianness.BIG
    config: Configuration = field(default_factory=Configuration)
    engine: Engine = Engine.TORCH
    logger.debug("Initialised")

    # braket specifics
    # TODO: include it in the configuration?
    _device: LocalSimulator = field(default_factory=LocalSimulator)

    def __post_init__(self) -> None:
        if self.is_remote:
            raise NotImplementedError("Braket backend does not support cloud execution yet")

    def circuit(self, circuit: QuantumCircuit) -> ConvertedCircuit:
        passes = self.config.transpilation_passes
        if passes is None:
            passes = default_passes

        original_circ = circuit
        if len(passes) > 0:
            circuit = transpile(*passes)(circuit)

        native = BraketCircuit(convert_block(circuit.block))
        return ConvertedCircuit(native=native, abstract=circuit, original=original_circ)

    def observable(self, obs: AbstractBlock, n_qubits: int = None) -> Any:
        if n_qubits is None:
            n_qubits = obs.n_qubits
        native = block_to_tensor(
            block=obs,
            values={},
            qubit_support=tuple([i for i in range(n_qubits)]),
            endianness=Endianness.BIG,
        ).squeeze(0)
        return ConvertedObservable(native=native, abstract=obs, original=obs)

    def run(
        self,
        circuit: ConvertedCircuit,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        """
        Execute the circuit and return a wavefunction in form of a statevector.

        Arguments:
            circuit: The circuit that is executed.
            param_values: Parameters of the circuit (after calling the embedding
                function on the user-facing parameters).
            state: Initial state.
            endianness: The endianness of the wave function.
        """

        if state is not None:
            raise NotImplementedError

        if self.is_remote:
            # handle here, or different backends?
            raise NotImplementedError

        # loop over all values in the batch
        results = []
        for vals in to_list_of_dicts(param_values):
            final_circuit = self.assign_parameters(circuit, vals)

            final_circuit.state_vector()  # set simulation type
            task = self._device.run(final_circuit, 0)
            results.append(task.result().values[0])
        states = torch.tensor(np.array(results))

        n_qubits = circuit.abstract.n_qubits
        if endianness != self.native_endianness and n_qubits > 1:
            from qadence.transpile import invert_endianness

            states = invert_endianness(states)
        return states

    def run_dm(
        self,
        circuit: ConvertedCircuit,
        noise: Noise,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        raise NotImplementedError

    def sample(
        self,
        circuit: ConvertedCircuit,
        param_values: dict[str, Tensor] = {},
        n_shots: int = 1,
        state: Tensor | None = None,
        noise: Noise | None = None,
        mitigation: Mitigations | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> list[Counter]:
        """Execute the circuit and return samples of the resulting wavefunction."""
        if state is not None:
            raise NotImplementedError("Braket cannot handle a custom initial state.")

        if n_shots < 1:
            raise ValueError("You can only call sample with n_shots>0.")

        if self.is_remote:
            # handle here, or different backends?
            raise NotImplementedError

        # loop over all values in the batch

        samples = []
        for vals in to_list_of_dicts(param_values):
            final_circuit = self.assign_parameters(circuit, vals)
            task = self._device.run(final_circuit, n_shots)
            samples.append(task.result().measurement_counts)
        if endianness != self.native_endianness:
            from qadence.transpile import invert_endianness

            samples = invert_endianness(samples)
        if noise is not None:
            samples = apply_noise(noise=noise, samples=samples)
        if mitigation is not None:
            logger.warning(
                "Mitigation protocol is deprecated. Use qadence-protocols instead.",
            )
            assert noise
            samples = apply_mitigation(noise=noise, mitigation=mitigation, samples=samples)
        return samples

    def expectation(
        self,
        circuit: ConvertedCircuit,
        observable: list[ConvertedObservable] | ConvertedObservable,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        measurement: Measurements | None = None,
        noise: Noise | None = None,
        mitigation: Mitigations | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        # Noise is ignored if measurement protocol is not provided.
        if noise is not None and measurement is None:
            logger.warning(
                f"Errors of type {noise} are not implemented for exact expectation yet. "
                "This is ignored for now."
            )
        # Do not flip endianness here because then we would have to reverse the observable
        wfs = self.run(circuit, param_values, state=state, endianness=Endianness.BIG)

        # TODO: Handle batching
        res = []
        observable = observable if isinstance(observable, list) else [observable]
        for wf in wfs:
            res.append([torch.vdot(wf, obs.native @ wf).real for obs in observable])

        return torch.tensor(res)

    def assign_parameters(
        self, circuit: ConvertedCircuit, param_values: dict[str, Tensor | float]
    ) -> BraketCircuit:
        """Assign numerical values to the circuit parameters."""
        if param_values is None:
            return circuit.native()

        params_copy = param_values.copy()
        pnames = [p.name for p in circuit.native.parameters]

        # account for fixed parameters
        for name in param_values.keys():
            if name not in pnames:
                params_copy.pop(name)

        # make sure that all the parameters are single floats
        # otherwise it won't be accepted by Braket
        native_params = promote_parameters(params_copy)

        # assign the parameters to the circuit
        assigned_circuit = circuit.native(**native_params)

        return assigned_circuit

    @staticmethod
    def _overlap(bras: Tensor, kets: Tensor) -> Tensor:
        return overlap_exact(bras, kets)

    @staticmethod
    def default_configuration() -> Configuration:
        return Configuration()



from __future__ import annotations

import os
from collections import Counter, OrderedDict
from dataclasses import asdict
from logging import getLogger
from pathlib import Path
from typing import Any, Callable, Optional, Sequence

import torch
from torch import Tensor, nn

from qadence.backend import (
    Backend,
    BackendConfiguration,
    BackendName,
    ConvertedCircuit,
    ConvertedObservable,
)
from qadence.backends.api import backend_factory, config_factory
from qadence.blocks.abstract import AbstractBlock
from qadence.blocks.utils import chain, unique_parameters
from qadence.circuit import QuantumCircuit
from qadence.engines.differentiable_backend import DifferentiableBackend
from qadence.measurements import Measurements
from qadence.mitigations import Mitigations
from qadence.noise import Noise
from qadence.parameters import Parameter
from qadence.types import DiffMode, Endianness

logger = getLogger(__name__)


class QuantumModel(nn.Module):
    """The central class of qadence that executes `QuantumCircuit`s and make them differentiable.

    This class should be used as base class for any new quantum model supported in the qadence
    framework for information on the implementation of custom models see
    [here](../tutorials/advanced_tutorials/custom-models.md).

    Example:
    ```python exec="on" source="material-block" result="json"
    import torch
    from qadence import QuantumModel, QuantumCircuit, RX, RY, Z, PI, chain, kron
    from qadence import FeatureParameter, VariationalParameter

    theta = VariationalParameter("theta")
    phi = FeatureParameter("phi")

    block = chain(
        kron(RX(0, theta), RY(1, theta)),
        kron(RX(0, phi), RY(1, phi)),
    )

    circuit = QuantumCircuit(2, block)

    observable = Z(0) + Z(1)

    model = QuantumModel(circuit, observable)
    values = {"phi": torch.tensor([PI, PI/2]), "theta": torch.tensor([PI, PI/2])}

    wf = model.run(values)
    xs = model.sample(values, n_shots=100)
    ex = model.expectation(values)
    print(wf)
    print(xs)
    print(ex)
    ```
    ```
    """

    backend: Backend | DifferentiableBackend
    embedding_fn: Callable
    _params: nn.ParameterDict
    _circuit: ConvertedCircuit
    _observable: list[ConvertedObservable] | None
    logger.debug("Initialised")

    def __init__(
        self,
        circuit: QuantumCircuit,
        observable: list[AbstractBlock] | AbstractBlock | None = None,
        backend: BackendName | str = BackendName.PYQTORCH,
        diff_mode: DiffMode = DiffMode.AD,
        measurement: Measurements | None = None,
        noise: Noise | None = None,
        mitigation: Mitigations | None = None,
        configuration: BackendConfiguration | dict | None = None,
    ):
        """Initialize a generic QuantumModel instance.

        Arguments:
            circuit: The circuit that is executed.
            observable: Optional observable(s) that are used only in the `expectation` method. You
                can also provide observables on the fly to the expectation call directly.
            backend: A backend for circuit execution.
            diff_mode: A differentiability mode. Parameter shift based modes work on all backends.
                AD based modes only on PyTorch based backends.
            measurement: Optional measurement protocol. If None, use
                exact expectation value with a statevector simulator.
            configuration: Configuration for the backend.
            noise: A noise model to use.

        Raises:
            ValueError: if the `diff_mode` argument is set to None
        """
        super().__init__()

        if not isinstance(circuit, QuantumCircuit):
            TypeError(
                f"The circuit should be of type '<class QuantumCircuit>'. Got {type(circuit)}."
            )

        if diff_mode is None:
            raise ValueError("`diff_mode` cannot be `None` in a `QuantumModel`.")

        self.backend = backend_factory(
            backend=backend, diff_mode=diff_mode, configuration=configuration
        )

        if isinstance(observable, list) or observable is None:
            observable = observable
        else:
            observable = [observable]

        def _is_feature_param(p: Parameter) -> bool:
            return not p.trainable and not p.is_number

        if observable is None:
            self.inputs = list(filter(_is_feature_param, circuit.unique_parameters))
        else:
            uparams = unique_parameters(chain(circuit.block, *observable))
            self.inputs = list(filter(_is_feature_param, uparams))

        conv = self.backend.convert(circuit, observable)
        self.embedding_fn = conv.embedding_fn
        self._circuit = conv.circuit
        self._observable = conv.observable
        self._backend_name = backend
        self._diff_mode = diff_mode
        self._measurement = measurement
        self._noise = noise
        self._mitigation = mitigation
        self._params = nn.ParameterDict(
            {
                str(key): nn.Parameter(val, requires_grad=val.requires_grad)
                for key, val in conv.params.items()
            }
        )

    @property
    def vparams(self) -> OrderedDict:
        """Variational parameters."""
        return OrderedDict({k: v.data for k, v in self._params.items() if v.requires_grad})

    @property
    def vals_vparams(self) -> Tensor:
        """Dictionary with parameters which are actually updated during optimization."""
        vals = torch.tensor([v for v in self._params.values() if v.requires_grad])
        vals.requires_grad = False
        return vals.flatten()

    @property
    def in_features(self) -> int:
        """Number of inputs."""
        return len(self.inputs)

    @property
    def out_features(self) -> int | None:
        """Number of outputs."""
        return 0 if self._observable is None else len(self._observable)

    @property
    def num_vparams(self) -> int:
        """The number of variational parameters."""
        return len(self.vals_vparams)

    def circuit(self, circuit: QuantumCircuit) -> ConvertedCircuit:
        """Get backend-converted circuit.

        Args:
            circuit: QuantumCircuit instance.

        Returns:
            Backend circuit.
        """
        return self.backend.circuit(circuit)

    def observable(self, observable: AbstractBlock, n_qubits: int) -> Any:
        """Get backend observable.

        Args:
            observable: Observable block.
            n_qubits: Number of qubits

        Returns:
            Backend observable.
        """
        return self.backend.observable(observable, n_qubits)

    def reset_vparams(self, values: Sequence) -> None:
        """Reset all the variational parameters with a given list of values."""
        current_vparams = OrderedDict({k: v for k, v in self._params.items() if v.requires_grad})

        assert (
            len(values) == self.num_vparams
        ), "Pass an iterable with the values of all variational parameters"
        for i, k in enumerate(current_vparams.keys()):
            current_vparams[k].data = torch.tensor([values[i]])

    def forward(self, *args: Any, **kwargs: Any) -> Tensor:
        """Calls run method with arguments.

        Returns:
            Tensor: A torch.Tensor representing output.
        """
        return self.run(*args, **kwargs)

    def run(
        self,
        values: dict[str, Tensor] = None,
        state: Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        r"""Run model.

        Given an input state $| \psi_0 \rangle$,
        a set of variational parameters $\vec{\theta}$
        and the unitary representation of the model $U(\vec{\theta})$
        we return $U(\vec{\theta}) | \psi_0 \rangle$.

        Arguments:
            values: Values dict which contains values for the parameters.
            state: Optional input state to apply model on.
            endianness: Storage convention for binary information.

        Returns:
            A torch.Tensor representing output.
        """
        if values is None:
            values = {}

        params = self.embedding_fn(self._params, values)

        return self.backend.run(self._circuit, params, state=state, endianness=endianness)

    def sample(
        self,
        values: dict[str, torch.Tensor] = {},
        n_shots: int = 1000,
        state: torch.Tensor | None = None,
        noise: Noise | None = None,
        mitigation: Mitigations | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> list[Counter]:
        """Obtain samples from model.

        Arguments:
            values: Values dict which contains values for the parameters.
            n_shots: Observable part of the expectation.
            state: Optional input state to apply model on.
            noise: A noise model to use.
            mitigation: A mitigation protocol to use.
            endianness: Storage convention for binary information.

        Returns:
            A list of Counter instances with the sample results.
        """
        params = self.embedding_fn(self._params, values)
        if noise is None:
            noise = self._noise
        if mitigation is None:
            mitigation = self._mitigation
        return self.backend.sample(
            self._circuit,
            params,
            n_shots=n_shots,
            state=state,
            noise=noise,
            mitigation=mitigation,
            endianness=endianness,
        )

    def expectation(
        self,
        values: dict[str, Tensor] = {},
        observable: list[ConvertedObservable] | ConvertedObservable | None = None,
        state: Optional[Tensor] = None,
        measurement: Measurements | None = None,
        noise: Noise | None = None,
        mitigation: Mitigations | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        r"""Compute expectation using the given backend.



        Given an input state $|\psi_0 \rangle$,
        a set of variational parameters $\vec{\theta}$
        and the unitary representation of the model $U(\vec{\theta})$
        we return $\langle \psi_0 | U(\vec{\theta}) | \psi_0 \rangle$.

        Arguments:
            values: Values dict which contains values for the parameters.
            observable: Observable part of the expectation.
            state: Optional input state.
            measurement: Optional measurement protocol. If None, use
                exact expectation value with a statevector simulator.
            noise: A noise model to use.
            mitigation: A mitigation protocol to use.
            endianness: Storage convention for binary information.

        Raises:
            ValueError: when no observable is set.

        Returns:
            A torch.Tensor of shape n_batches x n_obs
        """
        if observable is None:
            if self._observable is None:
                raise ValueError(
                    "Provide an AbstractBlock as the observable to compute expectation."
                    "Either pass a 'native_observable' directly to 'QuantumModel.expectation'"
                    "or pass a (non-native) '<class AbstractBlock>' to the 'QuantumModel.__init__'."
                )
            observable = self._observable

        params = self.embedding_fn(self._params, values)
        if measurement is None:
            measurement = self._measurement
        if noise is None:
            noise = self._noise
        else:
            self._noise = noise
        if mitigation is None:
            mitigation = self._mitigation
        return self.backend.expectation(
            circuit=self._circuit,
            observable=observable,
            param_values=params,
            state=state,
            measurement=measurement,
            noise=noise,
            mitigation=mitigation,
            endianness=endianness,
        )

    def overlap(self) -> Tensor:
        """Overlap of model.

        Raises:
            NotImplementedError: The overlap method is not implemented for this model.
        """
        raise NotImplementedError("The overlap method is not implemented for this model.")

    def _to_dict(self, save_params: bool = False) -> dict[str, Any]:
        """Convert QuantumModel to a dictionary for serialization.

        Arguments:
            save_params: Optionally save parameters. Defaults to False.

        Returns:
            The dictionary
        """
        d = dict()
        try:
            if isinstance(self._observable, list):
                abs_obs = [obs.abstract._to_dict() for obs in self._observable]
            else:
                abs_obs = [dict()]

            d = {
                "circuit": self._circuit.abstract._to_dict(),
                "observable": abs_obs,
                "backend": self._backend_name,
                "diff_mode": self._diff_mode,
                "measurement": (
                    self._measurement._to_dict() if self._measurement is not None else dict()
                ),
                "noise": self._noise._to_dict() if self._noise is not None else dict(),
                "backend_configuration": asdict(self.backend.backend.config),  # type: ignore
            }
            param_dict_conv = {}
            if save_params:
                param_dict_conv = {name: param.data for name, param in self._params.items()}
            d = {self.__class__.__name__: d, "param_dict": param_dict_conv}
            logger.debug(f"{self.__class__.__name__} serialized to {d}.")
        except Exception as e:
            logger.warning(f"Unable to serialize {self.__class__.__name__} due to {e}.")
        return d

    @classmethod
    def _from_dict(cls, d: dict, as_torch: bool = False) -> QuantumModel:
        """Initialize instance of QuantumModel from dictionary.

        Args:
            d: Dictionary.
            as_torch: Load parameters as torch tensors. Defaults to False.

        Returns:
            QuantumModel instance
        """
        from qadence.serialization import deserialize

        qm: QuantumModel
        try:
            qm_dict = d[cls.__name__]
            qm = cls(
                circuit=QuantumCircuit._from_dict(qm_dict["circuit"]),
                observable=(
                    None
                    if not isinstance(qm_dict["observable"], list)
                    else [deserialize(q_obs) for q_obs in qm_dict["observable"]]  # type: ignore[misc]
                ),
                backend=qm_dict["backend"],
                diff_mode=qm_dict["diff_mode"],
                measurement=Measurements._from_dict(qm_dict["measurement"]),
                noise=Noise._from_dict(qm_dict["noise"]),
                configuration=config_factory(qm_dict["backend"], qm_dict["backend_configuration"]),
            )

            if as_torch:
                conv_pd = torch.nn.ParameterDict()
                param_dict = d["param_dict"]
                for n, param in param_dict.items():
                    conv_pd[n] = torch.nn.Parameter(param)
                qm._params = conv_pd
            logger.debug(f"Initialized {cls.__name__} from {d}.")

        except Exception as e:
            logger.warning(f"Unable to deserialize object {d} to {cls.__name__} due to {e}.")

        return qm

    def save(
        self, folder: str | Path, file_name: str = "quantum_model.pt", save_params: bool = True
    ) -> None:
        """Save model.

        Arguments:
            folder: Folder where model is saved.
            file_name: File name for saving model. Defaults to "quantum_model.pt".
            save_params: Save parameters if True. Defaults to True.

        Raises:
            FileNotFoundError: If folder is not a directory.
        """
        if not os.path.isdir(folder):
            raise FileNotFoundError
        try:
            torch.save(self._to_dict(save_params), folder / Path(file_name))
        except Exception as e:
            logger.error(f"Unable to write QuantumModel to disk due to {e}")

    @classmethod
    def load(
        cls, file_path: str | Path, as_torch: bool = False, map_location: str | torch.device = "cpu"
    ) -> QuantumModel:
        """Load QuantumModel.

        Arguments:
            file_path: File path to load model from.
            as_torch: Load parameters as torch tensor. Defaults to False.
            map_location (str | torch.device, optional): Location for loading. Defaults to "cpu".

        Returns:
            QuantumModel from file_path.
        """
        qm_pt = {}
        if isinstance(file_path, str):
            file_path = Path(file_path)
        if os.path.isdir(file_path):
            from qadence.ml_tools.saveload import get_latest_checkpoint_name

            file_path = file_path / get_latest_checkpoint_name(file_path, "model")

        try:
            qm_pt = torch.load(file_path, map_location=map_location)
        except Exception as e:
            logger.error(f"Unable to load QuantumModel due to {e}")
        return cls._from_dict(qm_pt, as_torch)

    def assign_parameters(self, values: dict[str, Tensor]) -> Any:
        """Return the final, assigned circuit that is used in e.g. `backend.run`.

        Arguments:
            values: Values dict which contains values for the parameters.

        Returns:
            Final, assigned circuit that is used in e.g. `backend.run`
        """
        params = self.embedding_fn(self._params, values)
        return self.backend.assign_parameters(self._circuit, params)

    def to(self, *args: Any, **kwargs: Any) -> QuantumModel:
        """Conversion method for device or types.

        Returns:
            QuantumModel with conversions.
        """
        from pyqtorch import QuantumCircuit as PyQCircuit

        try:
            if isinstance(self._circuit.native, PyQCircuit):
                self._circuit.native = self._circuit.native.to(*args, **kwargs)
                if self._observable is not None:
                    if isinstance(self._observable, ConvertedObservable):
                        self._observable.native = self._observable.native.to(*args, **kwargs)
                    elif isinstance(self._observable, list):
                        for obs in self._observable:
                            obs.native = obs.native.to(*args, **kwargs)
                self._params = self._params.to(
                    device=self._circuit.native.device,
                    dtype=(
                        torch.float64
                        if self._circuit.native.dtype == torch.cdouble
                        else torch.float32
                    ),
                )
                logger.debug(f"Moved {self} to {args}, {kwargs}.")
            else:
                logger.debug("QuantumModel.to only supports pyqtorch.QuantumCircuits.")
        except Exception as e:
            logger.warning(f"Unable to move {self} to {args}, {kwargs} due to {e}.")
        return self

    @property
    def device(self) -> torch.device:
        """Get device.

        Returns:
            torch.device
        """
        return (
            self._circuit.native.device
            if self.backend.backend.name == "pyqtorch"  # type: ignore[union-attr]
            else torch.device("cpu")
        )


# Modules to be automatically added to the qadence namespace
__all__ = ["QuantumModel"]  # type: ignore



from __future__ import annotations

from qadence.backend import Backend, ConvertedCircuit, ConvertedObservable
from qadence.engines.differentiable_backend import (
    DifferentiableBackend as DifferentiableBackendInterface,
)
from qadence.engines.jax.differentiable_expectation import DifferentiableExpectation
from qadence.measurements import Measurements
from qadence.mitigations import Mitigations
from qadence.noise import Noise
from qadence.types import ArrayLike, DiffMode, Endianness, Engine, ParamDictType


class DifferentiableBackend(DifferentiableBackendInterface):
    """A class which wraps a QuantumBackend with the automatic differentation engine JAX.

    Arguments:
        backend: An instance of the QuantumBackend type perform execution.
        diff_mode: A differentiable mode supported by the differentiation engine.
        **psr_args: Arguments that will be passed on to `DifferentiableExpectation`.
    """

    def __init__(
        self,
        backend: Backend,
        diff_mode: DiffMode = DiffMode.AD,
        **psr_args: int | float | None,
    ) -> None:
        super().__init__(backend=backend, engine=Engine.JAX, diff_mode=diff_mode)
        self.psr_args = psr_args

    def expectation(
        self,
        circuit: ConvertedCircuit,
        observable: list[ConvertedObservable] | ConvertedObservable,
        param_values: ParamDictType = {},
        state: ArrayLike | None = None,
        measurement: Measurements | None = None,
        noise: Noise | None = None,
        mitigation: Mitigations | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> ArrayLike:
        """Compute the expectation value of the `circuit` with the given `observable`.

        Arguments:
            circuit: A converted circuit as returned by `backend.circuit`.
            observable: A converted observable as returned by `backend.observable`.
            param_values: _**Already embedded**_ parameters of the circuit. See
                [`embedding`][qadence.blocks.embedding.embedding] for more info.
            state: Initial state.
            measurement: Optional measurement protocol. If None, use
                exact expectation value with a statevector simulator.
            noise: A noise model to use.
            mitigation: The error mitigation to use.
            endianness: Endianness of the resulting bit strings.
        """
        observable = observable if isinstance(observable, list) else [observable]

        if self.diff_mode == DiffMode.AD:
            expectation = self.backend.expectation(circuit, observable, param_values, state)
        else:
            expectation = DifferentiableExpectation(
                backend=self.backend,
                circuit=circuit,
                observable=observable,
                param_values=param_values,
                state=state,
                measurement=measurement,
                noise=noise,
                mitigation=mitigation,
                endianness=endianness,
            ).psr()
        return expectation



from __future__ import annotations

from abc import abstractproperty
from dataclasses import dataclass
from typing import Tuple

import numpy as np
import torch
from rich.console import Console, RenderableType
from rich.tree import Tree
from sympy import Basic

from qadence.blocks.primitive import AbstractBlock
from qadence.parameters import Parameter, ParamMap, evaluate
from qadence.qubit_support import QubitSupport
from qadence.register import Register
from qadence.types import Interaction


@dataclass(eq=False, repr=False)
class AnalogBlock(AbstractBlock):
    @abstractproperty  # type: ignore[misc, override]
    def qubit_support(self) -> QubitSupport:
        pass

    @abstractproperty
    def duration(self) -> Parameter:
        pass

    def __ascii__(self, console: Console) -> RenderableType:
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        raise NotImplementedError

    @classmethod
    def _from_dict(cls, d: dict) -> AnalogBlock:
        raise NotImplementedError

    def _to_dict(self) -> dict:
        raise NotImplementedError

    @property
    def depth(self) -> int:
        raise NotImplementedError

    @property
    def n_qubits(self) -> int:
        if self.qubit_support.is_global:
            raise ValueError("Cannot compute number of qubits of a block with global support.")
        return max(self.qubit_support) + 1  # type: ignore[no-any-return]

    @property
    def n_supports(self) -> int:
        if self.qubit_support.is_global:
            raise ValueError("Cannot compute number of qubits of a block with global support.")
        return len(self.qubit_support)  # type: ignore[no-any-return]

    @property
    def eigenvalues_generator(self) -> torch.Tensor:
        msg = (
            "Eigenvalues of for generator of analog blocks can be computed via "
            "`add_background_hamiltonian(block, register).eigenvalues_generator`. "
        )
        raise NotImplementedError(msg)

    @property
    def eigenvalues(self) -> torch.Tensor:
        msg = (
            "Eigenvalues of analog blocks can be computed via "
            "`add_background_hamiltonian(block, register).eigenvalues`. "
        )
        raise NotImplementedError(msg)

    @property
    def _block_title(self) -> str:
        t = self.duration
        q = self.qubit_support
        s = f"{type(self).__name__}(t={evaluate(t)}, support={q})"

        if self.tag is not None:
            s += rf" \[tag: {self.tag}]"
        return s

    def compute_eigenvalues_generator(
        self,
        block: AbstractBlock,
        register: Register,
    ) -> torch.Tensor:
        # FIXME: Revisit analog blocks eigenvalues
        from qadence.analog import add_background_hamiltonian

        return add_background_hamiltonian(block, register).eigenvalues_generator  # type: ignore [union-attr]

    def dagger(self) -> AbstractBlock:
        raise NotImplementedError(
            f"Hermitian adjoint of block type {type(self)} is not implemented yet."
        )


@dataclass(eq=False, repr=False)
class InteractionBlock(AnalogBlock):
    """
    Free-evolution for the Hamiltonian interaction term of a register of qubits.

    In real interacting quantum devices, it means letting the system evolve freely according
    to the time-dependent Schrodinger equation. With emulators, this block is translated to an
    appropriate interaction Hamiltonian, for example, an Ising interaction

        Hᵢₙₜ = ∑ᵢⱼ C₆/rᵢⱼ⁶ nᵢnⱼ

    or an XY-interaction

        Hᵢₙₜ = ∑ᵢⱼ C₃/rⱼⱼ³ (XᵢXⱼ + ZᵢZⱼ)

    with `nᵢ = (1-Zᵢ)/2`.

    To construct, use the [`AnalogInteraction`][qadence.operations.AnalogInteraction] function.
    """

    _eigenvalues_generator: torch.Tensor | None = None

    parameters: ParamMap = ParamMap(duration=1000.0)  # ns
    qubit_support: QubitSupport = QubitSupport("global")

    add_pattern: bool = True

    @property
    def eigenvalues_generator(self) -> torch.Tensor | None:
        return self._eigenvalues_generator

    @eigenvalues_generator.setter
    def eigenvalues_generator(self, value: torch.Tensor) -> None:
        if not isinstance(value, torch.Tensor):
            value = torch.tensor(value)
        self._eigenvalues_generator = value

    @property
    def duration(self) -> Basic:
        return self.parameters.duration


@dataclass(eq=False, repr=False)
class ConstantAnalogRotation(AnalogBlock):
    """Implements a constant analog rotation with interaction dictated by the chosen Hamiltonian.

        H/h = ∑ᵢ(Ω/2 cos(φ)*Xᵢ - sin(φ)*Yᵢ - δnᵢ) + Hᵢₙₜ.

    To construct this block you can use of the following convenience wrappers:
    - The general rotation operation [`AnalogRot`][qadence.operations.AnalogRot]
    - Shorthands for rotatins around an axis:
      [`AnalogRX`][qadence.operations.AnalogRX],
      [`AnalogRY`][qadence.operations.AnalogRY],
      [`AnalogRZ`][qadence.operations.AnalogRZ]

    WARNING: do not use `ConstantAnalogRotation` with `alpha` as differentiable parameter - use
    the convenience wrappers mentioned above.
    """

    _eigenvalues_generator: torch.Tensor | None = None

    parameters: ParamMap = ParamMap(
        alpha=0.0,  # rad
        duration=1000.0,  # ns
        omega=0.0,  # rad/μs
        delta=0.0,  # rad/μs
        phase=0.0,  # rad
    )
    qubit_support: QubitSupport = QubitSupport("global")

    add_pattern: bool = True

    @property
    def _block_title(self) -> str:
        a = self.parameters.alpha
        t = self.parameters.duration
        q = self.qubit_support
        o = self.parameters.omega
        d = self.parameters.delta
        p = self.parameters.phase
        s = f"{type(self).__name__}(α={a}, t={t}, support={q}, Ω={o}, δ={d}, φ={p})"

        if self.tag is not None:
            s += rf" \[tag: {self.tag}]"
        return s

    @property
    def eigenvalues_generator(self) -> torch.Tensor:
        if self._eigenvalues_generator is None:
            raise ValueError(
                "Set ConstantAnalogRotation eigenvalues with compute_eigenvalues_generator method."
            )
        return self._eigenvalues_generator

    @eigenvalues_generator.setter
    def eigenvalues_generator(self, value: torch.Tensor) -> None:
        if not isinstance(value, torch.Tensor):
            value = torch.tensor(value)
        self._eigenvalues_generator = value

    @property
    def duration(self) -> Basic:
        return self.parameters.duration


####################################################################################################


# new, more strict versions of chain/kron blocks to make sure there are no gaps in composed blocks


@dataclass(eq=False, repr=False, init=False)
class AnalogComposite(AnalogBlock):
    blocks: Tuple[AnalogBlock, ...] = ()

    def __init__(self, blocks: Tuple[AnalogBlock, ...]):
        self.blocks = blocks
        # FIXME: add additional InteractionBlock if we have parameterized durations

    @property  # type: ignore[misc, override]
    def qubit_support(self) -> QubitSupport:
        return sum([b.qubit_support for b in self.blocks], start=QubitSupport())

    @abstractproperty
    def duration(self) -> Parameter:
        pass

    @property
    def _block_title(self) -> str:
        t = self.duration
        q = self.qubit_support
        s = f"{type(self).__name__}(t={t}, support={q})"

        if self.tag is not None:
            s += rf" \[tag: {self.tag}]"
        return s

    def __rich_tree__(self, tree: Tree = None) -> Tree:
        if tree is None:
            tree = Tree(self._block_title)
        else:
            tree = tree.add(self._block_title)
        for block in self.blocks:
            block.__rich_tree__(tree)
        return tree


@dataclass(eq=False, repr=False, init=False)
class AnalogChain(AnalogComposite):
    def __init__(self, blocks: Tuple[AnalogBlock, ...]):
        """A chain of analog blocks.

        Needed because analog blocks require
        stricter validation than the general `ChainBlock`.

        `AnalogChain`s can only be constructed from `AnalogKron` blocks or
        _**globally supported**_, primitive, analog blocks (like `InteractionBlock`s and
        `ConstantAnalogRotation`s).

        Automatically constructed by the [`chain`][qadence.blocks.utils.chain]
        function if only analog blocks are given.

        Example:
        ```python exec="on" source="material-block" result="json"
        from qadence import X, chain, AnalogInteraction

        b = chain(AnalogInteraction(200), AnalogInteraction(200))
        print(type(b))  # this is an `AnalogChain`

        b = chain(X(0), AnalogInteraction(200))
        print(type(b))  # this is a general `ChainBlock`
        ```
        """
        for b in blocks:
            if not (isinstance(b, AnalogKron) or b.qubit_support.is_global):
                raise ValueError("Only KronBlocks or global blocks can be chain'ed.")
        self.blocks = blocks

    @property
    def duration(self) -> Parameter:
        return Parameter(sum(evaluate(b.duration) for b in self.blocks))


@dataclass(eq=False, repr=False, init=False)
class AnalogKron(AnalogComposite):
    def __init__(self, blocks: Tuple[AnalogBlock, ...], interaction: Interaction = Interaction.NN):
        """Stack analog blocks vertically (i.e. in time).

        Needed because analog require
        stricter validation than the general `KronBlock`.

        `AnalogKron`s can only be constructed from _**non-global**_, analog blocks
        with the _**same duration**_.
        """
        if len(blocks) == 0:
            raise NotImplementedError("Empty KronBlocks not supported")

        self.blocks = blocks
        self.interaction = interaction

        qubit_support = QubitSupport()
        duration = blocks[0].duration
        for b in blocks:
            if not isinstance(b, AnalogBlock):
                raise ValueError("Can only kron `AnalgoBlock`s with other `AnalgoBlock`s.")

            if b.qubit_support == QubitSupport("global"):
                raise ValueError("Blocks with global support cannot be kron'ed.")

            if not qubit_support.is_disjoint(b.qubit_support):
                raise ValueError("Make sure blocks act on distinct qubits!")

            if not np.isclose(evaluate(duration), evaluate(b.duration)):
                raise ValueError("Kron'ed blocks have to have same duration.")

            qubit_support += b.qubit_support

        self.blocks = blocks

    @property
    def duration(self) -> Parameter:
        return self.blocks[0].duration


def chain(*args: AnalogBlock) -> AnalogChain:
    return AnalogChain(blocks=args)


def kron(*args: AnalogBlock) -> AnalogKron:
    return AnalogKron(blocks=args)



from __future__ import annotations

from typing import Sequence

import torch
from torch.nn import Module

from qadence import Noise
from qadence.blocks import (
    AbstractBlock,
)
from qadence.blocks.block_to_tensor import (
    block_to_tensor,
)
from qadence.utils import Endianness

from .config import Configuration


def convert_observable(
    block: AbstractBlock, n_qubits: int | None, config: Configuration = None
) -> Sequence[Module]:
    return [PulserObservable(block, n_qubits)]


class PulserObservable(Module):
    def __init__(self, block: AbstractBlock, n_qubits: int | None):
        super().__init__()
        self.block = block
        self.n_qubits = n_qubits if n_qubits else max(block.qubit_support) + 1

        if not self.block.is_parametric:
            block_mat = block_to_tensor(
                self.block, {}, qubit_support=tuple(i for i in range(self.n_qubits))
            ).squeeze(0)
            self.register_buffer("block_mat", block_mat)

    def forward(
        self,
        state: torch.Tensor,
        values: dict[str, torch.Tensor] = dict(),
        qubit_support: tuple | None = None,
        noise: Noise | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> torch.Tensor:
        if not self.block.is_parametric:
            block_mat = self.block_mat
        else:
            block_mat = block_to_tensor(
                self.block, values, qubit_support=qubit_support, endianness=endianness  # type: ignore [arg-type]  # noqa
            ).squeeze(0)
        if noise is None:  # Compute expectations for state vector.
            return torch.sum(torch.matmul(state, block_mat) * state.conj(), dim=1)
        else:  # Compute expectations for density matrices.
            return torch.trace(torch.matmul(block_mat, state))



from __future__ import annotations

import os
from pathlib import Path
from typing import Dict


class BaseTheme:
    name = ""
    background_color = ""
    color = ""
    fontname = "JetBrains Mono"
    fontsize = "20"
    primitive_node: dict[str, str] = {}
    fixed_parametric_node: dict[str, str] = {}
    feature_parametric_node: dict[str, str] = {}
    variational_parametric_node: dict[str, str] = {}
    hamevo_cluster: dict[str, str] = {}
    add_cluster: dict[str, str] = {}
    scale_cluster: dict[str, str] = {}

    @classmethod
    def get_graph_attr(self) -> Dict[str, str]:
        return {
            "bgcolor": self.background_color,
            "nodesep": "0.15",  # This defines the distance between wires
        }

    @classmethod
    def get_node_attr(self) -> Dict[str, str]:
        return {
            "color": self.color,
            "fontcolor": self.color,
            "fontname": self.fontname,
            "fontsize": self.fontsize,
            "width": "0.8",
            "height": "0.8",
            "style": "filled,rounded",
            "shape": "box",
            "penwidth": "2.7",
            "margin": "0.3,0.1",
        }

    @classmethod
    def get_edge_attr(self) -> Dict[str, str]:
        return {"color": self.color, "penwidth": "2.7"}

    @classmethod
    def get_cluster_attr(self) -> Dict[str, str]:
        return {
            "color": self.color,
            "fontcolor": self.color,
            "fontname": self.fontname,
            "fontsize": str(int(int(self.fontsize) * 2 / 3)),
            "labelloc": "t",
            "style": "rounded",
            "penwidth": "2.0",
        }

    @classmethod
    def get_primitive_node_attr(self) -> Dict[str, str]:
        attrs = self.get_node_attr()
        attrs.update(self.primitive_node)
        return attrs

    @classmethod
    def get_fixed_parametric_node_attr(self) -> Dict[str, str]:
        attrs = self.get_node_attr()
        attrs.update(self.fixed_parametric_node)
        return attrs

    @classmethod
    def get_feature_parametric_node_attr(self) -> Dict[str, str]:
        attrs = self.get_node_attr()
        attrs.update(self.feature_parametric_node)
        return attrs

    @classmethod
    def get_variational_parametric_node_attr(self) -> Dict[str, str]:
        attrs = self.get_node_attr()
        attrs.update(self.variational_parametric_node)
        return attrs

    @classmethod
    def get_add_cluster_attr(self) -> Dict[str, str]:
        attrs = self.get_cluster_attr()
        attrs.update(self.add_cluster)
        attrs["bgcolor"] = attrs["fillcolor"]
        return attrs

    @classmethod
    def get_scale_cluster_attr(self) -> Dict[str, str]:
        attrs = self.get_cluster_attr()
        attrs.update(self.scale_cluster)
        attrs["bgcolor"] = attrs["fillcolor"]
        return attrs

    @classmethod
    def get_hamevo_cluster_attr(self) -> Dict[str, str]:
        attrs = self.get_cluster_attr()
        attrs.update(self.hamevo_cluster)
        attrs["bgcolor"] = attrs["fillcolor"]
        return attrs

    @classmethod
    def load_measurement_icon(self) -> str:
        basedir = Path(os.path.abspath(os.path.dirname(__file__)))
        return os.path.join(basedir, "assets", self.name, "measurement.svg")


class Dark(BaseTheme):
    name = "dark"
    background_color = "black"
    color = "white"
    primitive_node = {"fillcolor": "#d03a2f", "color": "#f0c1be"}
    variational_parametric_node = {"fillcolor": "#3182bd", "color": "#afd3e9"}
    fixed_parametric_node = {"fillcolor": "#e6550d", "color": "#fed4b0"}
    feature_parametric_node = {"fillcolor": "#2d954d", "color": "#b2e8c2"}
    hamevo_cluster = {"fillcolor": "black", "color": "grey"}
    add_cluster = {"fillcolor": "black", "color": "grey"}
    scale_cluster = {"fillcolor": "black", "color": "grey"}


class Light(BaseTheme):
    name = "light"
    background_color = "white"
    color = "black"
    primitive_node = {"color": "#d03a2f", "fillcolor": "#f0c1be"}
    variational_parametric_node = {"color": "#3182bd", "fillcolor": "#afd3e9"}
    fixed_parametric_node = {"color": "#e6550d", "fillcolor": "#fed4b0"}
    feature_parametric_node = {"color": "#2d954d", "fillcolor": "#b2e8c2"}
    hamevo_cluster = {"color": "black", "fillcolor": "lightgrey"}
    add_cluster = {"color": "black", "fillcolor": "lightgrey"}
    scale_cluster = {"color": "black", "fillcolor": "lightgrey"}


class Black(BaseTheme):
    name = "black"
    background_color = "black"
    color = "white"
    primitive_node = {"color": "white", "fillcolor": "black"}
    variational_parametric_node = {"color": "white", "fillcolor": "black"}
    fixed_parametric_node = {"color": "white", "fillcolor": "black"}
    feature_parametric_node = {"color": "white", "fillcolor": "black"}
    hamevo_cluster = {"color": "white", "fillcolor": "black"}
    add_cluster = {"color": "white", "fillcolor": "black"}
    scale_cluster = {"color": "white", "fillcolor": "black"}


class White(BaseTheme):
    name = "white"
    background_color = "white"
    color = "black"
    primitive_node = {"color": "black", "fillcolor": "white"}
    variational_parametric_node = {"color": "black", "fillcolor": "white"}
    fixed_parametric_node = {"color": "black", "fillcolor": "white"}
    feature_parametric_node = {"color": "black", "fillcolor": "white"}
    hamevo_cluster = {"color": "black", "fillcolor": "white"}
    add_cluster = {"color": "black", "fillcolor": "white"}
    scale_cluster = {"color": "black", "fillcolor": "white"}



from __future__ import annotations

from dataclasses import dataclass
from logging import getLogger
from typing import Callable

from qadence.analog import add_background_hamiltonian
from qadence.backend import BackendConfiguration
from qadence.transpile import (
    blockfn_to_circfn,
    flatten,
    scale_primitive_blocks_only,
)

logger = getLogger(__name__)


def default_passes(config: Configuration) -> list[Callable]:
    passes: list = []

    # Replaces AnalogBlocks with respective HamEvo in the circuit block tree:
    passes.append(add_background_hamiltonian)

    # Flattens nested composed blocks:
    passes.append(lambda circ: blockfn_to_circfn(flatten)(circ))

    # Pushes block scales into the leaves of the block tree:
    passes.append(blockfn_to_circfn(scale_primitive_blocks_only))

    return passes


@dataclass
class Configuration(BackendConfiguration):
    pass



from __future__ import annotations

from .exceptions import NotPauliBlockError, NotSupportedError, QadenceException

# Modules to be automatically added to the qadence namespace
__all__ = ["exceptions"]



from __future__ import annotations

import importlib
from logging import getLogger
from typing import TypeVar

from qadence.backend import Backend, BackendConfiguration
from qadence.blocks.abstract import TAbstractBlock
from qadence.engines.differentiable_backend import DifferentiableBackend
from qadence.types import BackendName, DiffMode, Engine

BackendClsType = TypeVar("BackendClsType", bound=Backend)
EngineClsType = TypeVar("EngineClsType", bound=DifferentiableBackend)

logger = getLogger(__name__)


def import_config(backend_name: str | BackendName) -> BackendConfiguration:
    module_path = f"qadence.backends.{backend_name}.config"
    cfg: BackendConfiguration
    try:
        module = importlib.import_module(module_path)
        cfg = getattr(module, "Configuration")
    except (ModuleNotFoundError, ImportError) as e:
        raise type(e)(f"Failed to import backend config of {backend_name} due to {e}.") from e
    return cfg


def import_backend(backend_name: str | BackendName) -> Backend:
    module_path = f"qadence.backends.{backend_name}.backend"
    backend: Backend
    try:
        module = importlib.import_module(module_path)
    except (ModuleNotFoundError, ImportError) as e:
        # If backend is not in Qadence, search in extensions.
        module_path = f"qadence_extensions.backends.{backend_name}.backend"
        try:
            module = importlib.import_module(module_path)
        except (ModuleNotFoundError, ImportError) as e:
            raise type(e)
    backend = getattr(module, "Backend")
    return backend


def _available_backends() -> dict[BackendName, Backend]:
    """Return a dictionary of currently installed, native qadence backends."""
    res: dict[BackendName, Backend] = dict()
    for backend in BackendName.list():
        try:
            res[backend] = import_backend(backend)
        except (ModuleNotFoundError, ImportError) as e:
            raise type(e)(f"Failed to import backend {backend} due to {e}.") from e
    logger.debug(f"Found backends: {res.keys()}")
    return res


def import_engine(engine_name: str | Engine) -> DifferentiableBackend:
    module_path = f"qadence.engines.{engine_name}.differentiable_backend"
    engine: DifferentiableBackend
    try:
        module = importlib.import_module(module_path)
        engine = getattr(module, "DifferentiableBackend")
    except (ModuleNotFoundError, ImportError) as e:
        raise type(e)
    return engine


def _available_engines() -> dict[Engine, DifferentiableBackend]:
    """Return a dictionary of currently installed, native qadence engines."""
    res: dict[Engine, DifferentiableBackend] = dict()
    for engine in Engine.list():
        try:
            res[engine] = import_engine(engine)
        except (ModuleNotFoundError, ImportError) as e:
            raise type(e)(f"Failed to import engine {engine} due to {e}.") from e
    logger.debug(f"Found engines: {res.keys()}")
    return res


def _supported_gates(backend_name: str) -> list[TAbstractBlock]:
    """Return a list of supported gates for the queried backend 'name'."""
    from qadence import operations

    backend_name = BackendName(backend_name)  # Validate backend name.
    module_path = f"qadence.backends.{backend_name}"

    try:
        module = importlib.import_module(module_path)
    except (ModuleNotFoundError, ImportError) as e:
        raise type(e)(f"Failed to import backend module for {backend_name} due to {e}.") from e
    _supported_gates = getattr(module, "supported_gates")
    return [getattr(operations, gate) for gate in _supported_gates]


def _gpsr_fns() -> dict:
    """Fallback function for native Qadence GPSR functions if extensions is not present."""
    # avoid circular import
    from qadence.backends.gpsr import general_psr

    return {DiffMode.GPSR: general_psr}


def _validate_diff_mode(backend: Backend, diff_mode: DiffMode) -> None:
    """Fallback function for native Qadence diff_mode if extensions is not present."""
    if not backend.supports_ad and diff_mode == DiffMode.AD:
        raise TypeError(f"Backend {backend.name} does not support diff_mode {DiffMode.AD}.")
    elif not backend.supports_adjoint and diff_mode == DiffMode.ADJOINT:
        raise TypeError(f"Backend {backend.name} does not support diff_mode {DiffMode.ADJOINT}.")


def _set_backend_config(backend: Backend, diff_mode: DiffMode) -> None:
    """Fallback function for native Qadence backends if extensions is not present.

    Args:
        backend (Backend): A backend for execution.
        diff_mode (DiffMode): A differentiation mode.
    """

    _validate_diff_mode(backend, diff_mode)

    # (1) When using PSR with any backend or (2) we use the backends Pulser or Braket,
    # we have to use gate-level parameters

    # We can use expression-level parameters for AD.
    if backend.name == BackendName.PYQTORCH:
        backend.config.use_single_qubit_composition = True
        backend.config._use_gate_params = diff_mode != "ad"
    else:
        backend.config._use_gate_params = True


# if proprietary qadence_plus is available import the
# right function since more backends are supported
try:
    module = importlib.import_module("qadence_extensions.extensions")
    available_backends = getattr(module, "available_backends")
    available_engines = getattr(module, "available_engines")
    supported_gates = getattr(module, "supported_gates")
    get_gpsr_fns = getattr(module, "gpsr_fns")
    set_backend_config = getattr(module, "set_backend_config")
except ModuleNotFoundError:
    available_backends = _available_backends
    available_engines = _available_engines
    supported_gates = _supported_gates
    get_gpsr_fns = _gpsr_fns
    set_backend_config = _set_backend_config



from __future__ import annotations

from .addressing import AddressingPattern
from .device import IdealDevice, RealisticDevice, RydbergDevice
from .parse_analog import add_background_hamiltonian

__all__ = ["RydbergDevice", "IdealDevice", "RealisticDevice", "AddressingPattern"]



from __future__ import annotations

import logging
from warnings import warn


def get_logger(name: str) -> logging.Logger:
    warn(
        '"get_logger" will be deprected soon.\
         Please use "get_script_logger" instead.',
        DeprecationWarning,
    )
    return logging.getLogger(name)


def get_script_logger(name: str = "") -> logging.Logger:
    return logging.getLogger(f"script.{name}")



from __future__ import annotations

import torch
from torch import Tensor

from qadence.backend import Backend
from qadence.backends.pyqtorch import Backend as PyQBackend
from qadence.blocks import AbstractBlock
from qadence.blocks.utils import unroll_block_with_scaling
from qadence.circuit import QuantumCircuit
from qadence.engines.differentiable_backend import DifferentiableBackend
from qadence.measurements.utils import iterate_pauli_decomposition
from qadence.noise import Noise
from qadence.utils import Endianness


def compute_expectation(
    circuit: QuantumCircuit,
    observables: list[AbstractBlock],
    param_values: dict,
    options: dict,
    state: Tensor | None = None,
    backend: Backend | DifferentiableBackend = PyQBackend(),
    noise: Noise | None = None,
    endianness: Endianness = Endianness.BIG,
) -> Tensor:
    """Basic tomography protocol with rotations.

    Given a circuit and a list of observables, apply basic tomography protocol to estimate
    the expectation values.

    Args:
        circuit (QuantumCircuit): a circuit to prepare the state.
        observables (list[AbstractBlock]): a list of observables
            to estimate the expectation values from.
        param_values (dict): a dict of values to substitute the
            symbolic parameters for.
        options (dict): a dict of options for the measurement protocol.
            Here, shadow_size (int), accuracy (float) and confidence (float) are supported.
        state (Tensor | None): an initial input state.
        backend_name (BackendName): a backend name to retrieve computations from.
        noise: A noise model to use.
        endianness: Endianness of the observable estimate.
    """
    if not isinstance(observables, list):
        raise TypeError(
            "Observables must be of type <class 'List[AbstractBlock]'>. Got {}.".format(
                type(observables)
            )
        )
    n_shots = options.get("n_shots")
    if n_shots is None:
        raise KeyError("Tomography protocol requires a 'n_shots' kwarg of type 'int'.")
    estimated_values = []
    for observable in observables:
        pauli_decomposition = unroll_block_with_scaling(observable)
        estimated_values.append(
            iterate_pauli_decomposition(
                circuit=circuit,
                param_values=param_values,
                pauli_decomposition=pauli_decomposition,
                n_shots=n_shots,
                state=state,
                backend=backend,
                noise=noise,
                endianness=endianness,
            )
        )
    return torch.transpose(torch.vstack(estimated_values), 1, 0)



from __future__ import annotations

import numpy as np
from sympy import Basic

from qadence.backend import BackendConfiguration
from qadence.blocks import chain, kron
from qadence.blocks.abstract import AbstractBlock
from qadence.blocks.composite import ChainBlock, KronBlock
from qadence.blocks.utils import add
from qadence.circuit import QuantumCircuit
from qadence.constructors import (
    analog_feature_map,
    feature_map,
    hamiltonian_factory,
    identity_initialized_ansatz,
    rydberg_feature_map,
    rydberg_hea,
    rydberg_tower_feature_map,
)
from qadence.constructors.ansatze import hea_digital, hea_sDAQC
from qadence.constructors.hamiltonians import ObservableConfig, TDetuning
from qadence.measurements import Measurements
from qadence.noise import Noise
from qadence.operations import CNOT, RX, RY, I, N, Z
from qadence.parameters import Parameter
from qadence.register import Register
from qadence.types import (
    AnsatzType,
    BackendName,
    DiffMode,
    InputDiffMode,
    Interaction,
    MultivariateStrategy,
    ObservableTransform,
    ReuploadScaling,
    Strategy,
    TParameter,
)

from .config import AnsatzConfig, FeatureMapConfig
from .models import QNN


def _create_support_arrays(
    num_qubits: int,
    num_features: int,
    multivariate_strategy: str,
) -> list[tuple[int, ...]]:
    """
    Create the support arrays for the digital feature map.

    Args:
        num_qubits (int): The number of qubits.
        num_features (int): The number of features.
        multivariate_strategy (str): The multivariate encoding strategy.
            Either 'series' or 'parallel'.

    Returns:
        list[tuple[int, ...]]: The list of support arrays. ith element of the list is the support
            array for the ith feature.

    Raises:
        ValueError: If the number of features is greater than the number of qubits
            with parallel encoding. Not possible to encode these features in parallel.
        ValueError: If the multivariate strategy is not 'series' or 'parallel'.
    """
    if multivariate_strategy == "series":
        return [tuple(range(num_qubits)) for i in range(num_features)]
    elif multivariate_strategy == "parallel":
        if num_features <= num_qubits:
            return [tuple(x.tolist()) for x in np.array_split(np.arange(num_qubits), num_features)]
        else:
            raise ValueError(
                f"Number of features {num_features} must be less than or equal to the number of \
                qubits {num_qubits}. if the features are to be encoded is parallely."
            )
    else:
        raise ValueError(
            f"Invalid encoding strategy {multivariate_strategy} provided. Only 'series' or \
                'parallel' are allowed."
        )


def _encode_features_series_digital(
    register: int | Register,
    config: FeatureMapConfig,
) -> list[AbstractBlock]:
    """
    Encode the features in series using digital feature map.

    Args:
        register (int | Register): The number of qubits or the register object.
        config (FeatureMapConfig): The feature map configuration.

    Returns:
        list[AbstractBlock]: The list of digital feature map blocks.
    """
    num_qubits = register if isinstance(register, int) else register.n_qubits

    support_arrays_list = _create_support_arrays(
        num_qubits=num_qubits,
        num_features=config.num_features,
        multivariate_strategy=config.multivariate_strategy,
    )

    support_arrays = {
        key: support for key, support in zip(config.inputs, support_arrays_list)  # type: ignore[union-attr, arg-type]
    }

    num_uploads = {key: value + 1 for key, value in config.num_repeats.items()}  # type: ignore[union-attr]

    if config.param_prefix is None:
        param_prefixes = {feature: None for feature in config.inputs}  # type: ignore[union-attr]
    else:
        param_prefixes = {feature: f"{config.param_prefix}_{feature}" for feature in config.inputs}  # type: ignore

    fm_blocks: list[AbstractBlock] = []

    for i in range(max(num_uploads.values())):
        for feature in config.inputs:  # type: ignore[union-attr]
            if num_uploads[feature] > 0:
                fm_blocks.append(
                    feature_map(
                        num_qubits,
                        support=support_arrays[feature],
                        param=feature,
                        op=config.operation,  # type: ignore[arg-type]
                        fm_type=config.basis_set[feature],  # type: ignore[arg-type, index]
                        reupload_scaling=config.reupload_scaling[feature],  # type: ignore[arg-type, index]
                        feature_range=config.feature_range[feature],  # type: ignore[arg-type, index]
                        target_range=config.target_range[feature],  # type: ignore[arg-type, index]
                        param_prefix=param_prefixes[feature],
                    )
                )
            num_uploads[feature] -= 1

    return fm_blocks


def _encode_features_parallel_digital(
    register: int | Register,
    config: FeatureMapConfig,
) -> list[AbstractBlock]:
    """
    Encode the features in parallel using digital feature map.

    Args:
        register (int | Register): The number of qubits or the register object.
        config (FeatureMapConfig): The feature map configuration.

    Returns:
        list[AbstractBlock]: The list of digital feature map blocks.
    """
    num_qubits = register if isinstance(register, int) else register.n_qubits

    support_arrays_list = _create_support_arrays(
        num_qubits=num_qubits,
        num_features=config.num_features,
        multivariate_strategy=config.multivariate_strategy,
    )

    support_arrays = {
        key: support for key, support in zip(config.inputs, support_arrays_list)  # type: ignore[union-attr, arg-type]
    }

    num_uploads = {key: value + 1 for key, value in config.num_repeats.items()}  # type: ignore[union-attr]

    if config.param_prefix is None:
        param_prefixes = {feature: None for feature in config.inputs}  # type: ignore[union-attr]
    else:
        param_prefixes = {feature: f"{config.param_prefix}_{feature}" for feature in config.inputs}  # type: ignore

    fm_blocks: list[AbstractBlock] = []

    for i in range(max(num_uploads.values())):
        fm_layer = []
        for feature in config.inputs:  # type: ignore[union-attr]
            if num_uploads[feature] > 0:
                fm_layer.append(
                    feature_map(
                        len(support_arrays[feature]),
                        support=support_arrays[feature],
                        param=feature,
                        op=config.operation,  # type: ignore[arg-type]
                        fm_type=config.basis_set[feature],  # type: ignore[index]
                        reupload_scaling=config.reupload_scaling[feature],  # type: ignore[index]
                        feature_range=config.feature_range[feature],  # type: ignore[arg-type, index]
                        target_range=config.target_range[feature],  # type: ignore[arg-type, index]
                        param_prefix=param_prefixes[feature],
                    )
                )
            num_uploads[feature] -= 1

        fm_blocks.append(kron(*fm_layer))

    return fm_blocks


def _create_digital_fm(
    register: int | Register,
    config: FeatureMapConfig,
) -> list[AbstractBlock]:
    """
    Create the digital feature map.

    Args:
        register (int | Register): The number of qubits or the register object.
        config (FeatureMapConfig): The feature map configuration.

    Returns:
        list[AbstractBlock]: The list of digital feature map blocks.

    Raises:
        ValueError: If the encoding strategy is invalid. Only 'series' or 'parallel' are allowed.
    """
    if config.multivariate_strategy == MultivariateStrategy.SERIES:
        fm_blocks = _encode_features_series_digital(register, config)
    elif config.multivariate_strategy == MultivariateStrategy.PARALLEL:
        fm_blocks = _encode_features_parallel_digital(register, config)
    else:
        raise ValueError(
            f"Invalid encoding strategy {config.multivariate_strategy} provided. Only 'series' or \
                'parallel' are allowed."
        )

    return fm_blocks


def _create_analog_fm(
    register: int | Register,
    config: FeatureMapConfig,
) -> list[AbstractBlock]:
    """
    Create the analog feature map.

    Args:
        register (int | Register): The number of qubits or the register object.
        config (FeatureMapConfig): The feature map configuration.

    Returns:
        list[AbstractBlock]: The list of analog feature map blocks.
    """

    num_uploads = {key: value + 1 for key, value in config.num_repeats.items()}  # type: ignore[union-attr]

    fm_blocks: list[AbstractBlock] = []

    for i in range(max(num_uploads.values())):
        for feature in config.inputs:  # type: ignore[union-attr]
            if num_uploads[feature] > 0:
                fm_blocks.append(
                    analog_feature_map(
                        param=feature,
                        op=config.operation,  # type: ignore[arg-type]
                        fm_type=config.basis_set[feature],  # type: ignore[arg-type, index]
                        reupload_scaling=config.reupload_scaling[feature],  # type: ignore[arg-type, index]
                        feature_range=config.feature_range[feature],  # type: ignore[arg-type, index]
                        target_range=config.target_range[feature],  # type: ignore[arg-type, index]
                    )
                )
            num_uploads[feature] -= 1

    return fm_blocks


def _encode_feature_rydberg(
    num_qubits: int,
    param: str,
    reupload_scaling: ReuploadScaling,
) -> AbstractBlock:
    """
    Encode features using a Rydberg feature map.

    Args:
        num_qubits (int): The number of qubits to encode the features on.
        param (str): The parameter prefix to use for the feature map parameter names.
            reupload_scaling (ReuploadScaling): The scaling strategy for reuploads.

    Returns:
        The Rydberg feature map.

    Raises:
        NotImplementedError: If the reupload scaling strategy is not implemented.
            Only `ReuploadScaling.CONSTANT` and `ReuploadScaling.TOWER` are supported.
    """
    if reupload_scaling == ReuploadScaling.CONSTANT:
        return rydberg_feature_map(n_qubits=num_qubits, param=param)

    elif reupload_scaling == ReuploadScaling.TOWER:
        return rydberg_tower_feature_map(n_qubits=num_qubits, param=param)

    else:
        raise NotImplementedError(f"Rydberg feature map not implemented for {reupload_scaling}")


def _create_rydberg_fm(
    register: int | Register,
    config: FeatureMapConfig,
) -> list[AbstractBlock]:
    """
    Create a Rydberg feature map for the given configuration.

    Args:
        register (int | Register): The number of qubits or the register to apply the feature map to.
        config (FeatureMapConfig): The configuration for the feature map.

    Returns:
        list: A list of Rydberg feature map blocks.
    """
    num_qubits = register if isinstance(register, int) else register.n_qubits

    num_uploads = {key: value + 1 for key, value in config.num_repeats.items()}  # type: ignore[union-attr]

    fm_blocks = []

    for i in range(max(num_uploads.values())):
        for feature in config.inputs:  # type: ignore[union-attr]
            if num_uploads[feature] > 0:
                fm_blocks.append(
                    _encode_feature_rydberg(
                        num_qubits=num_qubits,
                        param=feature,
                        reupload_scaling=config.reupload_scaling[feature],  # type: ignore[arg-type, index]
                    )
                )
            num_uploads[feature] -= 1

    return fm_blocks


def create_fm_blocks(
    register: int | Register,
    config: FeatureMapConfig,
) -> list[AbstractBlock]:
    """
    Create a list of feature map blocks based on the given configuration.

    In case of series encoding or even parallel encoding with data reuploads,
    the outputs is a list of blocks that still need to be interleaved with non
    commuting blocks.

    Args:
        register (int | Register): The number of qubits or the register.
        config (FeatureMapConfig): The configuration for the feature map.

    Returns:
        list[AbstractBlock]: A list of feature map blocks.

    Raises:
        ValueError: If the feature map strategy is not 'digital', 'analog' or 'rydberg'.
    """
    if config.feature_map_strategy == Strategy.DIGITAL:
        return _create_digital_fm(register=register, config=config)
    elif config.feature_map_strategy == Strategy.ANALOG:
        return _create_analog_fm(register=register, config=config)
    elif config.feature_map_strategy == Strategy.RYDBERG:
        return _create_rydberg_fm(register=register, config=config)
    else:
        raise NotImplementedError(
            f"Feature map not implemented for strategy {config.feature_map_strategy}. \
            Only 'digital', 'analog' or 'rydberg' allowed."
        )


def _ansatz_layer(
    register: int | Register,
    ansatz_config: AnsatzConfig,
    index: int,
) -> AbstractBlock:
    """
    Create a layer of the ansatz based on the configuration.

    Args:
        register (int | Register): The number of qubits or the register.
        ansatz_config (AnsatzConfig): The configuration for the ansatz.
        index (int): The index of the layer.

    Returns:
        AbstractBlock: The layer of the ansatz.
    """
    new_config = AnsatzConfig(
        depth=1,
        ansatz_type=ansatz_config.ansatz_type,
        ansatz_strategy=ansatz_config.ansatz_strategy,
        strategy_args=ansatz_config.strategy_args,
        param_prefix=f"fm_{index}",
    )

    return create_ansatz(register=register, config=new_config)


def _create_iia_digital(
    num_qubits: int,
    config: AnsatzConfig,
) -> AbstractBlock:
    """
    Create the Digital Identity Initialized Ansatz based on the configuration.

    Args:
        num_qubits (int): The number of qubits.
        config (AnsatzConfig): The configuration for the ansatz.

    Returns:
        AbstractBlock: The Identity Initialized Ansatz.
    """
    operations = config.strategy_args.get("operations", [RX, RY])
    entangler = config.strategy_args.get("entangler", CNOT)
    periodic = config.strategy_args.get("periodic", False)

    return identity_initialized_ansatz(
        n_qubits=num_qubits,
        depth=config.depth,
        param_prefix=config.param_prefix,
        strategy=Strategy.DIGITAL,
        rotations=operations,
        entangler=entangler,
        periodic=periodic,
    )


def _create_iia_sdaqc(
    num_qubits: int,
    config: AnsatzConfig,
) -> AbstractBlock:
    """
    Create the SDAQC Identity Initialized Ansatz based on the configuration.

    Args:
        num_qubits (int): The number of qubits.
        config (AnsatzConfig): The configuration for the ansatz.

    Returns:
        AbstractBlock: The SDAQC Identity Initialized Ansatz.
    """
    operations = config.strategy_args.get("operations", [RX, RY])
    entangler = config.strategy_args.get("entangler", CNOT)
    periodic = config.strategy_args.get("periodic", False)

    return identity_initialized_ansatz(
        n_qubits=num_qubits,
        depth=config.depth,
        param_prefix=config.param_prefix,
        strategy=Strategy.SDAQC,
        rotations=operations,
        entangler=entangler,
        periodic=periodic,
    )


def _create_iia(
    num_qubits: int,
    config: AnsatzConfig,
) -> AbstractBlock:
    """
    Create the Identity Initialized Ansatz based on the configuration.

    Args:
        num_qubits (int): The number of qubits.
        config (AnsatzConfig): The configuration for the ansatz.

    Returns:
        AbstractBlock: The Identity Initialized Ansatz.

    Raises:
        ValueError: If the ansatz strategy is not supported. Only 'digital' and 'sdaqc' are allowed.
    """
    if config.ansatz_strategy == Strategy.DIGITAL:
        return _create_iia_digital(num_qubits=num_qubits, config=config)
    elif config.ansatz_strategy == Strategy.SDAQC:
        return _create_iia_sdaqc(num_qubits=num_qubits, config=config)
    else:
        raise ValueError(
            f"Invalid ansatz strategy {config.ansatz_strategy} provided. Only 'digital', 'sdaqc', \
            allowed for IIA."
        )


def _create_hea_digital(num_qubits: int, config: AnsatzConfig) -> AbstractBlock:
    """
    Create the Digital Hardware Efficient Ansatz based on the configuration.

    Args:
        num_qubits (int): The number of qubits.
        config (AnsatzConfig): The configuration for the ansatz.

    Returns:
        AbstractBlock: The Digital Hardware Efficient Ansatz.
    """
    operations = config.strategy_args.get("rotations", [RX, RY, RX])
    entangler = config.strategy_args.get("entangler", CNOT)
    periodic = config.strategy_args.get("periodic", False)

    return hea_digital(
        n_qubits=num_qubits,
        depth=config.depth,
        param_prefix=config.param_prefix,
        operations=operations,
        entangler=entangler,
        periodic=periodic,
    )


def _create_hea_sdaqc(num_qubits: int, config: AnsatzConfig) -> AbstractBlock:
    """
    Create the SDAQC Hardware Efficient Ansatz based on the configuration.

    Args:
        num_qubits (int): The number of qubits.
        config (AnsatzConfig): The configuration for the ansatz.

    Returns:
        AbstractBlock: The SDAQC Hardware Efficient Ansatz.
    """
    operations = config.strategy_args.get("rotations", [RX, RY, RX])
    entangler = config.strategy_args.get(
        "entangler", hamiltonian_factory(num_qubits, interaction=Interaction.NN)
    )

    return hea_sDAQC(
        n_qubits=num_qubits,
        depth=config.depth,
        param_prefix=config.param_prefix,
        operations=operations,
        entangler=entangler,
    )


def _create_hea_rydberg(
    register: int | Register,
    config: AnsatzConfig,
) -> AbstractBlock:
    """
    Create the Rydberg Hardware Efficient Ansatz based on the configuration.

    Args:
        register (int | Register): The number of qubits or the register object.
        config (AnsatzConfig): The configuration for the ansatz.

    Returns:
        AbstractBlock: The Rydberg Hardware Efficient Ansatz.
    """
    register = register if isinstance(register, Register) else Register.circle(n_qubits=register)

    addressable_detuning = config.strategy_args.get("addressable_detuning", True)
    addressable_drive = config.strategy_args.get("addressable_drive", False)
    tunable_phase = config.strategy_args.get("tunable_phase", False)

    return rydberg_hea(
        register=register,
        n_layers=config.depth,
        addressable_detuning=addressable_detuning,
        addressable_drive=addressable_drive,
        tunable_phase=tunable_phase,
        additional_prefix=config.param_prefix,
    )


def _create_hea_ansatz(
    register: int | Register,
    config: AnsatzConfig,
) -> AbstractBlock:
    """
    Create the Hardware Efficient Ansatz based on the configuration.

    Args:
        register (int | Register): The number of qubits or the register to create the ansatz for.
        config (AnsatzConfig): The configuration for the ansatz.

    Returns:
        AbstractBlock: The hardware efficient ansatz block.

    Raises:
        ValueError: If the ansatz strategy is not 'digital', 'sdaqc', or 'rydberg'.
    """
    num_qubits = register if isinstance(register, int) else register.n_qubits

    if config.ansatz_strategy == Strategy.DIGITAL:
        return _create_hea_digital(num_qubits=num_qubits, config=config)
    elif config.ansatz_strategy == Strategy.SDAQC:
        return _create_hea_sdaqc(num_qubits=num_qubits, config=config)
    elif config.ansatz_strategy == Strategy.RYDBERG:
        return _create_hea_rydberg(register=register, config=config)
    else:
        raise ValueError(
            f"Invalid ansatz strategy {config.ansatz_strategy} provided. Only 'digital', 'sdaqc', \
                and 'rydberg' allowed"
        )


def create_ansatz(
    register: int | Register,
    config: AnsatzConfig,
) -> AbstractBlock:
    """
    Create the ansatz based on the configuration.

    Args:
        register (int | Register): Number of qubits or a register object.
        config (AnsatzConfig): Configuration for the ansatz.

    Returns:
        AbstractBlock: The ansatz block.

    Raises:
        NotImplementedError: If the ansatz type is not implemented.
    """
    num_qubits = register if isinstance(register, int) else register.n_qubits

    if config.ansatz_type == AnsatzType.IIA:
        return _create_iia(num_qubits=num_qubits, config=config)
    elif config.ansatz_type == AnsatzType.HEA:
        return _create_hea_ansatz(register=register, config=config)
    else:
        raise NotImplementedError(
            f"Ansatz of type {config.ansatz_type} not implemented yet. Only 'hea' and\
                'iia' available."
        )


def _interleave_ansatz_in_fm(
    register: int | Register,
    fm_blocks: list[AbstractBlock],
    ansatz_config: AnsatzConfig,
) -> ChainBlock:
    """
    Interleave the ansatz layers in between the feature map layers.

    Args:
        register (int | Register): Number of qubits or a register object.
        fm_blocks (list[AbstractBlock]): List of feature map blocks.
        ansatz_config (AnsatzConfig): Ansatz configuration.

    Returns:
        ChainBlock: A block containing feature map layers with interleaved ansatz layers.
    """
    full_fm = []
    for idx, block in enumerate(fm_blocks):
        full_fm.append(block)
        if idx + 1 < len(fm_blocks):
            full_fm.append(_ansatz_layer(register, ansatz_config, idx))

    return chain(*full_fm)


def load_observable_transformations(config: ObservableConfig) -> tuple[Parameter, Parameter]:
    """
    Get the observable shifting and scaling factors.

    Args:
        config (ObservableConfig): Observable configuration.

    Returns:
        tuple[float, float]: The observable shifting and scaling factors.
    """
    shift = config.shift
    scale = config.scale
    if config.trainable_transform is not None:
        shift = Parameter(name=shift, trainable=config.trainable_transform)
        scale = Parameter(name=scale, trainable=config.trainable_transform)
    else:
        shift = Parameter(shift)
        scale = Parameter(scale)
    return scale, shift


ObservableTransformMap = {
    ObservableTransform.RANGE: lambda detuning, scale, shift: (shift, shift - scale)
    if detuning is N
    else (0.5 * (shift - scale), 0.5 * (scale + shift)),
    ObservableTransform.SCALE: lambda _, scale, shift: (scale, shift),
}


def _global_identity(register: int | Register) -> KronBlock:
    """Create a global identity block."""
    return kron(
        *[I(i) for i in range(register if isinstance(register, int) else register.n_qubits)]
    )


def observable_from_config(
    register: int | Register,
    config: ObservableConfig,
) -> AbstractBlock:
    """
    Create an observable block.

    Args:
        register (int | Register): Number of qubits or a register object.
        config (ObservableConfig): Observable configuration.

    Returns:
        AbstractBlock: The observable block.
    """
    scale, shift = load_observable_transformations(config)
    return create_observable(register, config.detuning, scale, shift, config.transformation_type)


def create_observable(
    register: int | Register,
    detuning: TDetuning = Z,
    scale: TParameter | None = None,
    shift: TParameter | None = None,
    transformation_type: ObservableTransform = ObservableTransform.NONE,  # type: ignore[assignment]
) -> AbstractBlock:
    """
    Create an observable block.

    Args:
        register (int | Register): Number of qubits or a register object.
        detuning: The type of detuning.
        scale: A parameter for the scale.
        shift: A parameter for the shift.

    Returns:
        AbstractBlock: The observable block.
    """
    if transformation_type == ObservableTransform.RANGE:
        scale, shift = ObservableTransformMap[transformation_type](detuning, scale, shift)  # type: ignore[index]
    shifting_term = shift * _global_identity(register)  # type: ignore[operator]
    detuning_hamiltonian = scale * hamiltonian_factory(  # type: ignore[operator]
        register=register,
        detuning=detuning,
    )
    return add(shifting_term, detuning_hamiltonian)


def build_qnn_from_configs(
    register: int | Register,
    observable_config: ObservableConfig | list[ObservableConfig],
    fm_config: FeatureMapConfig = FeatureMapConfig(),
    ansatz_config: AnsatzConfig = AnsatzConfig(),
    backend: BackendName = BackendName.PYQTORCH,
    diff_mode: DiffMode = DiffMode.AD,
    measurement: Measurements | None = None,
    noise: Noise | None = None,
    configuration: BackendConfiguration | dict | None = None,
    input_diff_mode: InputDiffMode | str = InputDiffMode.AD,
) -> QNN:
    """
    Build a QNN model.

    Args:
        register (int | Register): Number of qubits or a register object.
        observable_config (ObservableConfig | list[ObservableConfig]): Observable configuration(s).
        fm_config (FeatureMapConfig): Feature map configuration.
        ansatz_config (AnsatzConfig): Ansatz configuration.
        backend (BackendName): The chosen quantum backend.
        diff_mode (DiffMode): The differentiation engine to use. Choices are
            'gpsr' or 'ad'.
        measurement (Measurements): Optional measurement protocol. If None,
            use exact expectation value with a statevector simulator.
        noise (Noise): A noise model to use.
        configuration (BackendConfiguration | dict): Optional backend configuration.
        input_diff_mode (InputDiffMode): The differentiation mode for the input tensor.

    Returns:
        QNN: A QNN model.
    """
    blocks: list[AbstractBlock] = []
    inputs: list[Basic | str] | None = None

    if fm_config.num_features > 0:
        fm_blocks = create_fm_blocks(register=register, config=fm_config)
        full_fm = _interleave_ansatz_in_fm(
            register=register,
            fm_blocks=fm_blocks,
            ansatz_config=ansatz_config,
        )
        inputs = fm_config.inputs
        blocks.append(full_fm)

    blocks.append(create_ansatz(register=register, config=ansatz_config))

    circ = QuantumCircuit(register, *blocks)

    observable: AbstractBlock | list[AbstractBlock] = (
        [observable_from_config(register=register, config=cfg) for cfg in observable_config]
        if isinstance(observable_config, list)
        else observable_from_config(register=register, config=observable_config)
    )

    ufa = QNN(
        circ,
        observable,
        inputs=inputs,
        backend=backend,
        diff_mode=diff_mode,
        measurement=measurement,
        noise=noise,
        configuration=configuration,
        input_diff_mode=input_diff_mode,
    )

    return ufa



from __future__ import annotations

from dataclasses import dataclass, fields

from qadence.analog import AddressingPattern
from qadence.types import PI, DeviceType, Interaction


@dataclass(frozen=True, eq=True)
class RydbergDevice:
    """
    Dataclass for interacting Rydberg atoms under an Hamiltonian:

    H = ∑_i [Ω/2 * (cos(φ) * Xᵢ - sin(φ) * Yᵢ) - δ * N_i] + H_int,

    where:

    H_int = ∑_(j<i) (C_6 / R**6) * (N_i @ N_j) for the NN interaction;

    H_int = ∑_(j<i) (C_3 / R**3) * ((X_i @ X_j) + (Y_i @ Y_j)) for the XY interaction;
    """

    interaction: Interaction = Interaction.NN
    """Defines the interaction Hamiltonian."""

    rydberg_level: int = 60
    """Rydberg level affecting the value of C_6."""

    coeff_xy: float = 3700.00
    """Value of C_3."""

    max_detuning: float = 2 * PI * 4
    """Maximum value of the detuning δ."""

    max_amp: float = 2 * PI * 3
    """Maximum value of the amplitude Ω."""

    pattern: AddressingPattern | None = None
    """Semi-local addressing pattern configuration."""

    type: DeviceType = DeviceType.IDEALIZED
    """DeviceType.IDEALIZED or REALISTIC to convert to the Pulser backend."""

    def __post_init__(self) -> None:
        # FIXME: Currently not supporting custom interaction functions.
        if self.interaction not in [Interaction.NN, Interaction.XY]:
            raise KeyError(
                "RydbergDevice currently only supports Interaction.NN or Interaction.XY."
            )

    def _to_dict(self) -> dict:
        device_dict = {}
        for field in fields(self):
            if field.name != "pattern":
                device_dict[field.name] = getattr(self, field.name)
            else:
                device_dict[field.name] = (
                    self.pattern._to_dict() if self.pattern is not None else {}
                )
        return device_dict

    @classmethod
    def _from_dict(cls, d: dict) -> RydbergDevice:
        pattern = AddressingPattern._from_dict(d["pattern"])
        d["pattern"] = pattern
        return cls(**d)


def IdealDevice(pattern: AddressingPattern | None = None) -> RydbergDevice:
    return RydbergDevice(
        pattern=pattern,
        type=DeviceType.IDEALIZED,
    )


def RealisticDevice(pattern: AddressingPattern | None = None) -> RydbergDevice:
    return RydbergDevice(
        pattern=pattern,
        type=DeviceType.REALISTIC,
    )



from __future__ import annotations

from dataclasses import dataclass
from logging import getLogger
from typing import Callable

from pyqtorch.utils import SolverType

from qadence.analog import add_background_hamiltonian
from qadence.backend import BackendConfiguration
from qadence.transpile import (
    blockfn_to_circfn,
    chain_single_qubit_ops,
    flatten,
    scale_primitive_blocks_only,
)
from qadence.types import AlgoHEvo

logger = getLogger(__name__)


def default_passes(config: Configuration) -> list[Callable]:
    passes: list = []

    # Replaces AnalogBlocks with respective HamEvo in the circuit block tree:
    passes.append(add_background_hamiltonian)

    if config.use_single_qubit_composition:
        # Composes chains of single-qubit gates into a single unitary before applying to the state:
        passes.append(lambda circ: blockfn_to_circfn(chain_single_qubit_ops)(circ))
    else:
        # Flattens nested composed blocks:
        passes.append(lambda circ: blockfn_to_circfn(flatten)(circ))

    # Pushes block scales into the leaves of the block tree:
    passes.append(blockfn_to_circfn(scale_primitive_blocks_only))

    return passes


@dataclass
class Configuration(BackendConfiguration):
    algo_hevo: AlgoHEvo = AlgoHEvo.EXP
    """Determine which kind of Hamiltonian evolution algorithm to use."""

    ode_solver: SolverType = SolverType.DP5_SE
    """Determine which ODE solver to use for time-dependent blocks."""

    n_steps_hevo: int = 100
    """Default number of steps for the Hamiltonian evolution."""

    use_gradient_checkpointing: bool = False
    """Use gradient checkpointing.

    Recommended for higher-order optimization tasks.
    """

    use_single_qubit_composition: bool = False
    """Composes chains of single qubit gates into a single matmul if possible."""

    loop_expectation: bool = False
    """When computing batches of expectation values, only allocate one wavefunction.

    Loop over the batch of parameters to only allocate a single wavefunction at any given time.
    """



from __future__ import annotations

from collections import Counter
from math import log2
from typing import Callable, Sequence

import numpy as np
import pyqtorch as pyq
import torch
from numpy.typing import ArrayLike
from pyqtorch.apply import apply_operator
from pyqtorch.parametric import Parametric as PyQParametric
from torch import (
    Tensor,
    cat,
    complex64,
    complex128,
    mean,
    no_grad,
    rand,
)

from qadence.types import Endianness, ParamDictType
from qadence.utils import int_to_basis, is_qadence_shape

FINITE_DIFF_EPS = 1e-06
# Dict of NumPy dtype -> torch dtype (when the correspondence exists)
numpy_to_torch_dtype_dict = {
    np.bool_: torch.bool,
    np.uint8: torch.uint8,
    np.int8: torch.int8,
    np.int16: torch.int16,
    np.int32: torch.int32,
    np.int64: torch.int64,
    np.float16: torch.float16,
    np.float32: torch.float32,
    np.float64: torch.float64,
    np.complex64: torch.complex64,
    np.complex128: torch.complex128,
    int: torch.int64,
    float: torch.float64,
    complex: torch.complex128,
}


def param_dict(keys: Sequence[str], values: Sequence[Tensor]) -> dict[str, Tensor]:
    return {key: val for key, val in zip(keys, values)}


def numpy_to_tensor(
    x: np.ndarray,
    device: torch.device = torch.device("cpu"),
    dtype: torch.dtype = torch.complex128,
    requires_grad: bool = False,
) -> Tensor:
    """This only copies the numpy array if device or dtype are different than the ones of x."""
    return torch.as_tensor(x, dtype=dtype, device=device).requires_grad_(requires_grad)


def promote_to_tensor(
    x: Tensor | np.ndarray | float,
    dtype: torch.dtype = torch.complex128,
    requires_grad: bool = True,
) -> Tensor:
    """Convert the given type into a torch.Tensor."""
    if isinstance(x, float):
        return torch.tensor([[x]], dtype=dtype, requires_grad=requires_grad)
    elif isinstance(x, np.ndarray):
        return numpy_to_tensor(
            x, dtype=numpy_to_torch_dtype_dict.get(x.dtype), requires_grad=requires_grad
        )
    elif isinstance(x, Tensor):
        return x.requires_grad_(requires_grad)
    else:
        raise ValueError(f"Don't know how to promote {type(x)} to Tensor")


# FIXME: Not being used, maybe remove in v1.0.0
def count_bitstrings(sample: Tensor, endianness: Endianness = Endianness.BIG) -> Counter:
    # Convert to a tensor of integers.
    n_qubits = sample.size()[1]
    base = torch.ones(n_qubits, dtype=torch.int64) * 2
    powers_of_2 = torch.pow(base, reversed(torch.arange(n_qubits)))
    int_tensor = torch.matmul(sample, powers_of_2)
    # Count occurences of integers.
    count_int = torch.bincount(int_tensor)
    # Return a Counter for non-empty bitstring counts.
    return Counter(
        {
            int_to_basis(k=k, n_qubits=n_qubits, endianness=endianness): count.item()
            for k, count in enumerate(count_int)
            if count > 0
        }
    )


def to_list_of_dicts(param_values: ParamDictType) -> list[ParamDictType]:
    if not param_values:
        return [param_values]

    max_batch_size = max(p.size()[0] for p in param_values.values() if isinstance(p, Tensor))
    batched_values = {
        k: (v if v.size()[0] == max_batch_size else v.repeat(max_batch_size, 1))
        for k, v in param_values.items()
        if isinstance(v, Tensor)
    }

    return [{k: v[i] for k, v in batched_values.items()} for i in range(max_batch_size)]


def pyqify(state: Tensor, n_qubits: int = None) -> ArrayLike:
    """Convert a state of shape (batch_size, 2**n_qubits) to [2] * n_qubits + [batch_size]."""
    if n_qubits is None:
        n_qubits = int(log2(state.shape[1]))
    if len(state.shape) != 2 or (state.shape[1] != 2**n_qubits):
        raise ValueError(
            "The initial state must be composed of tensors/arrays of size "
            f"(batch_size, 2**n_qubits). Found: {state.shape = }."
        )
    return state.T.reshape([2] * n_qubits + [state.shape[0]])


def unpyqify(state: Tensor) -> Tensor:
    """Convert a state of shape [2] * n_qubits + [batch_size] to (batch_size, 2**n_qubits)."""
    return torch.flatten(state, start_dim=0, end_dim=-2).t()


def is_pyq_shape(state: Tensor, n_qubits: int) -> bool:
    return state.size()[:-1] == [2] * n_qubits  # type: ignore[no-any-return]


def validate_state(state: Tensor, n_qubits: int) -> None:
    """Check if a custom initial state conforms to the qadence or the pyqtorch format."""
    if state.dtype not in [complex128, complex64]:
        raise TypeError(f"Expected complex dtype, got {state.dtype}")
    elif len(state.size()) < 2:
        raise ValueError(f"Invalid state shape. Got {state.shape}")
    elif not is_qadence_shape(state, n_qubits) and not is_pyq_shape(state, n_qubits):
        raise ValueError(
            f"Allowed formats for custom initial state are:\
                  (1) Qadence shape: (batch_size, 2**n_qubits)\
                  (2) Pyqtorch shape: (2 * n_qubits + [batch_size])\
                  Found: {state.size() = }"
        )


def infer_batchsize(param_values: dict[str, Tensor] = None) -> int:
    """Infer the batch_size through the length of the parameter tensors."""
    try:
        return (
            max(
                [
                    len(tensor_or_dict)
                    for tensor_or_dict in param_values.values()
                    if isinstance(tensor_or_dict, Tensor)
                ]
            )
            if param_values
            else 1
        )
    except Exception:
        return 1


# The following functions can be used to compute potentially higher order gradients using pyqtorch's
# native 'jacobian' methods.


def finitediff(
    f: Callable,
    x: Tensor,
    derivative_indices: tuple[int, ...],
    eps: float = None,
) -> Tensor:
    """
    Compute the finite difference of a function at a point.

    Args:
        f: The function to differentiate.
        x: Input of size `(batch_size, input_size)`.
        derivative_indices: Which *input* to differentiate (i.e. which variable x[:,i])
        eps: finite difference spacing (uses `torch.finfo(x.dtype).eps ** (1 / (2 + order))`
            as default)

    Returns:
        (Tensor): The finite difference of the function at the point `x`.
    """

    if eps is None:
        order = len(derivative_indices)
        eps = torch.finfo(x.dtype).eps ** (1 / (2 + order))

    # compute derivative direction vector(s)
    eps = torch.as_tensor(eps, dtype=x.dtype)
    _eps = 1 / eps  # type: ignore[operator]
    ev = torch.zeros_like(x)
    i = derivative_indices[0]
    ev[:, i] += eps

    # recursive finite differencing for higher order than 3 / mixed derivatives
    if len(derivative_indices) > 3 or len(set(derivative_indices)) > 1:
        di = derivative_indices[1:]
        return (finitediff(f, x + ev, di) - finitediff(f, x - ev, di)) * _eps / 2
    elif len(derivative_indices) == 3:
        return (f(x + 2 * ev) - 2 * f(x + ev) + 2 * f(x - ev) - f(x - 2 * ev)) * _eps**3 / 2
    elif len(derivative_indices) == 2:
        return (f(x + ev) + f(x - ev) - 2 * f(x)) * _eps**2
    elif len(derivative_indices) == 1:
        return (f(x + ev) - f(x - ev)) * _eps / 2
    else:
        raise ValueError("If you see this error there is a bug in the `finitediff` function.")


def finitediff_sampling(
    f: Callable, x: Tensor, eps: float = FINITE_DIFF_EPS, num_samples: int = 10
) -> Tensor:
    def _finitediff(val: Tensor) -> Tensor:
        return (f(x + val) - f(x - val)) / (2 * val)  # type: ignore

    with no_grad():
        return mean(cat([_finitediff(val) for val in rand(1) for _ in range(num_samples)]))


def dydx(
    jacobian: Tensor,
    qubit_support: tuple,
    out_state: Tensor,
    projected_state: Tensor,
) -> Tensor:
    return 2 * pyq.inner_prod(
        projected_state,
        apply_operator(
            state=out_state,
            operator=jacobian,
            qubits=qubit_support,
        ).real,
    )


def dydxx(
    op: PyQParametric,
    values: dict[str, Tensor],
    out_state: Tensor,
    projected_state: Tensor,
) -> Tensor:
    return 2 * finitediff_sampling(
        lambda val: dydx(
            op.jacobian({op.param_name: val}), op.qubit_support, out_state, projected_state
        ),
        values[op.param_name],
    )
